
# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from contextlib import suppress
from httpx import AsyncClient

from langchain_core.language_models.base import BaseLanguageModel

from neuro_san.internals.run_context.langchain.llms.llm_policy import LlmPolicy


class OpenAILlmPolicy(LlmPolicy):
    """
    LlmPolicy implementation for OpenAI.

    OpenAI's BaseLanguageModel implementations do allow us to pass in a web client
    as an argument, so this implementation takes advantage of the create_client()
    method to do that. Worth noting that where many other implementations might care about
    the llm reference, because of our create_client() implementation, we do not.
    """

    def __init__(self, llm: BaseLanguageModel = None):
        """
        Constructor.
        """
        super().__init__()

        self.http_client: AsyncClient = None

        # Not doing lazy type resolution here just for type hints.
        # Save that for create_client(), where it's meatier.
        self.async_openai_client: Any = None

    def create_client(self, config: Dict[str, Any]) -> Any:
        """
        Creates the web client to used by a BaseLanguageModel to be
        constructed in the future.  Neuro SAN infrastructures prefers that this
        be an asynchronous client, however we realize some BaseLanguageModels
        do not support that (even though they should!).

        Implementations should retain any references to state that needs to be cleaned up
        in the delete_resources() method.

        :param config: The fully specified llm config
        :return: The web client that accesses the LLM.
                By default this is None, as many BaseLanguageModels
                do not allow a web client to be passed in as an arg.
        """
        # OpenAI is the one chat class that we do not require any extra installs.
        # This is what we want to work out of the box.
        # Nevertheless, have it go through the same lazy-loading resolver rigamarole as the others.

        # pylint: disable=invalid-name
        AsyncOpenAI = self.resolver.resolve_class_in_module("AsyncOpenAI",
                                                            module_name="openai",
                                                            install_if_missing="langchain-openai")

        self.create_http_client(config)

        self.async_openai_client = AsyncOpenAI(
            api_key=self.get_value_or_env(config, "openai_api_key", "OPENAI_API_KEY"),
            base_url=self.get_value_or_env(config, "openai_api_base", "OPENAI_API_BASE"),
            organization=self.get_value_or_env(config, "openai_organization", "OPENAI_ORG_ID"),
            timeout=config.get("request_timeout"),
            max_retries=config.get("max_retries"),
            http_client=self.http_client
        )

        # We retain the async_openai_client reference, but we hand back this reach-in
        # to pass to the BaseLanguageModel constructor.
        return self.async_openai_client.chat.completions

    def create_http_client(self, config: Dict[str, Any]):
        """
        Creates the http client from the given config.

        :param config: The fully specified llm config
        """
        # Our run-time model resource here is httpx client which we need to control directly:
        openai_proxy: str = self.get_value_or_env(config, "openai_proxy", "OPENAI_PROXY")
        request_timeout: int = config.get("request_timeout")
        self.http_client = AsyncClient(proxy=openai_proxy, timeout=request_timeout)

    def create_llm(self, config: Dict[str, Any], model_name: str, client: Any) -> BaseLanguageModel:
        """
        Create a BaseLanguageModel instance from the fully-specified llm config
        for the llm class that the implementation supports.  Chat models are usually
        per-provider, where the specific model itself is an argument to its constructor.

        :param config: The fully specified llm config
        :param model_name: The name of the model
        :param client: The web client to use (if any)
        :return: A BaseLanguageModel (can be Chat or LLM)
        """
        # OpenAI is the one chat class that we do not require any extra installs.
        # This is what we want to work out of the box.
        # Nevertheless, have it go through the same lazy-loading resolver rigamarole as the others.

        # pylint: disable=invalid-name
        ChatOpenAI = self.resolver.resolve_class_in_module("ChatOpenAI",
                                                           module_name="langchain_openai.chat_models.base",
                                                           install_if_missing="langchain-openai")

        # Now construct LLM chat model we will be using:
        llm = ChatOpenAI(
            async_client=client,
            model_name=model_name,
            temperature=config.get("temperature"),

            # This next group of params should always be None when we have async_client
            openai_api_key=self.get_value_or_env(config, "openai_api_key",
                                                 "OPENAI_API_KEY", client),
            openai_api_base=self.get_value_or_env(config, "openai_api_base",
                                                  "OPENAI_API_BASE", client),
            openai_organization=self.get_value_or_env(config, "openai_organization",
                                                      "OPENAI_ORG_ID", client),
            openai_proxy=self.get_value_or_env(config, "openai_organization",
                                               "OPENAI_PROXY", client),
            request_timeout=self.get_value_or_env(config, "request_timeout", None, client),
            max_retries=self.get_value_or_env(config, "max_retries", None, client),

            presence_penalty=config.get("presence_penalty"),
            frequency_penalty=config.get("frequency_penalty"),
            seed=config.get("seed"),
            logprobs=config.get("logprobs"),
            top_logprobs=config.get("top_logprobs"),
            logit_bias=config.get("logit_bias"),
            streaming=True,  # streaming is always on. Without it token counting will not work.
            n=1,  # n is always 1.  neuro-san will only ever consider one chat completion.
            top_p=config.get("top_p"),
            max_tokens=config.get("max_tokens"),  # This is always for output
            tiktoken_model_name=config.get("tiktoken_model_name"),
            stop=config.get("stop"),

            # The following three parameters are for reasoning models only.
            reasoning=config.get("reasoning"),
            reasoning_effort=config.get("reasoning_effort"),
            verbosity=config.get("verbosity"),

            # If omitted, this defaults to the global verbose value,
            # accessible via langchain_core.globals.get_verbose():
            # https://github.com/langchain-ai/langchain/blob/master/libs/core/langchain_core/globals.py#L53
            #
            # However, accessing the global verbose value during concurrent initialization
            # can trigger the following warning:
            #
            # UserWarning: Importing verbose from langchain root module is no longer supported.
            # Please use langchain.globals.set_verbose() / langchain.globals.get_verbose() instead.
            # old_verbose = langchain.verbose
            #
            # To prevent this, we explicitly set verbose=False here (which matches the default
            # global verbose value) so that the warning is never triggered.
            verbose=False,

            # Set stream_usage to True in order to get token counting chunks.
            stream_usage=True
        )

        return llm

    async def delete_resources(self):
        """
        Release the run-time resources used by the instance.
        """
        self.async_openai_client = None

        if self.http_client is not None:
            with suppress(Exception):
                await self.http_client.aclose()

        self.http_client = None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from types import ModuleType

from os import environ

from neuro_san.internals.authorization.interfaces.abstract_authorizer import AbstractAuthorizer
from neuro_san.internals.authorization.interfaces.authorizer import Authorizer
from neuro_san.internals.authorization.openfga.open_fga_store_cache import OpenFgaStoreCache


class OpenFgaAuthorizer(AbstractAuthorizer):
    """
    AbstractAuthorizer implementation for Open FGA ("Fine Grained Authorization").
    """

    def __init__(self, fga_client: Any = None):
        """
        Constructor

        :param fga_client: A pre-initialized OpenFgaClient
        """
        super().__init__()

        # Lazy importing of openfga_sdk so extra dependency can be optional
        self.openfga_sdk: ModuleType = self.resolver.resolve_class_in_module(class_name=None,
                                                                             module_name="openfga_sdk",
                                                                             install_if_missing="openfga-sdk")

        debug_auth: str = environ.get("AGENT_DEBUG_AUTH")
        self.debug: bool = debug_auth is not None and len(debug_auth) > 0 and debug_auth != "false"
        self.fail_on_unauthorized: bool = environ.get("AGENT_DEBUG_AUTH") == "hard"

        # Note: we don't initialize the client because constructors cannot be async
        # This is what we use aenter() and aexit() for.
        self.fga_client: self.openfga_sdk.client.client.OpenFgaClient = fga_client

    async def __aenter__(self) -> Authorizer:
        """
        Opens a scoped session with an Authorizer.
        """
        if self.fga_client is not None:
            await self.fga_client.close()

        # Open a new client
        self.fga_client = await OpenFgaStoreCache.get_client()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
        """
        Closes a scoped session with an Authorizer.
        """
        await self.fga_client.close()
        self.fga_client = None

    async def authorize(self, actor: Dict[str, Any], action: str, resource: Dict[str, Any]) -> bool:
        """
        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      is seeking permission.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param action: The action for which the user is asking permission for.
                      Most often this is one of the strings "create", "read", "update" or "delete".
        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: True if the actor is allowed to take the requested action on the resource.
                 False otherwise.
        """

        # Guilty until proven innocent
        authorized: bool = False

        # Create some local versions of args we might modify
        use_actor: Dict[str, Any] = actor
        use_action: str = action
        use_resource: Dict[str, Any] = resource

        # Fix case where values are Enum types.
        # OpenFGA needs everything to be JSON serializable, and the Enum types we have are not.
        if not isinstance(action, str):
            use_action = action.value
        if not isinstance(resource.get("type"), str):
            use_resource = {
                "type": resource.get("type").value,
                "id": resource.get("id")
            }

        # Useful in debugging
        if self.debug:
            self.logger.info("authorize(%s, %s, %s:%s)", use_actor, use_action,
                             use_resource.get("type"), use_resource.get("id"))

        # Use classes from the lazily imported module to avoid extra required dependencies
        # pylint: disable=invalid-name
        ClientCheckRequest = self.openfga_sdk.client.models.check_request.ClientCheckRequest
        CheckResponse = self.openfga_sdk.models.check_response.CheckResponse

        # Prepare a request to see if the server can tell us the answer.
        check_request = ClientCheckRequest(user=f"{use_actor.get('type')}:{use_actor.get('id')}",
                                           relation=use_action,
                                           object=f"{use_resource.get('type')}:{use_resource.get('id')}")

        # No async with here, as that would close the client
        check_response: CheckResponse = await self.fga_client.check(check_request)
        authorized = check_response.allowed

        if not authorized:
            message: str = f"Actor: {actor}   action: {action}   resource: {resource}"
            if self.debug:
                self.logger.info(message)
                self.logger.info("authorized is %s", authorized)

            if self.fail_on_unauthorized:
                # Exception useful when trying to figure out where authorization
                # is being made in testing.
                raise ValueError(message)

        return authorized

    # pylint: disable=too-many-locals
    async def list(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> List[str]:
        """
        Return a list of resource ids that the actor has the given relation to,
        as per the graph specified by the authorization model.

        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      entity's relation should be checked.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user's permissions will be checked.
                     Most often this is one of the strings from the Permission enum.

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            # Note: "id" is not specified. We want a list of these returned.
                        }
        :return: A list of resource ids that the actor has the given relation with.
                 An empty return list implies that the actor has access to no objects
                 of the given resource type.
        """
        # Initialize a return value
        ids: List[str] = []

        if resource.get("id") is not None:
            # We are looking for a specific id. Faster through authorize()
            if self.debug:
                self.logger.info("using authorize() for list()")
            authorized: bool = self.authorize(actor, relation, resource)
            if authorized:
                ids.append(str(resource.get("id")))
            return ids

        # Formulate the user specification for the request
        actor_type: str = actor.get("type", "")
        actor_id: str = actor.get("id", "")
        request_user: str = None
        if len(actor_type) > 0 or len(actor_id) > 0:
            request_user = f"{actor_type}:{actor_id}"

        # Formulate the object/resource specification for the request
        resource_type: str = resource.get("type", "")

        if self.debug:
            self.logger.info("list(%s, %s, %s:%s)", actor_id, relation, resource_type,  resource.get("id"))

        # Use classes from the lazily imported module to avoid extra required dependencies
        # pylint: disable=invalid-name
        ClientListObjectsRequest = self.openfga_sdk.client.models.list_objects_request.ClientListObjectsRequest
        ListObjectsResponse = self.openfga_sdk.models.list_objects_response.ListObjectsResponse

        # Make the API call
        options: Dict[str, Any] = {}
        body = ClientListObjectsRequest(user=request_user,
                                        relation=relation,
                                        type=resource_type)

        # No async with here, as that would close the client
        response: ListObjectsResponse = await self.fga_client.list_objects(body, options)

        for one_object in response.objects:
            # Results come in the format of a single string "<type>:<identifier>"
            one_id: str = one_object.split(":")[1]
            ids.append(one_id)

        if self.debug:
            self.logger.info("list length is %d", len(ids))

        return ids

    # pylint: disable=too-many-locals
    async def query(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> List[str]:
        """
        Instead of a boolean answer from authorize() above, this method gives a list
        of resources of the given resource type (in the dict) that the actor has the
        *direct* given relation to.  This does not take authorization policy graphs
        into account.

        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      will be permitted.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user will be permitted.

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: A list of relations (which can be None or empty) that the actor
                has the given relation with.
        """
        # Formulate the user specification for the request
        actor_type: str = ""
        actor_id: str = ""
        if actor is not None:
            actor_type = actor.get("type", "")
            actor_id = actor.get("id", "")

        request_user: str = None
        if len(actor_type) > 0 or len(actor_id) > 0:
            request_user = f"{actor_type}:{actor_id}"

        # Formulate the object/resource specification for the request
        resource_type: str = resource.get("type", "")
        resource_id: str = resource.get("id", "")
        request_object: str = None
        if len(resource_type) > 0 or len(resource_id) > 0:
            request_object = f"{resource_type}:{resource_id}"

        # Use classes from the lazily imported module to avoid extra required dependencies
        # pylint: disable=invalid-name
        ReadRequestTupleKey = self.openfga_sdk.models.read_request_tuple_key.ReadRequestTupleKey
        ReadResponse = self.openfga_sdk.models.read_response.ReadResponse
        TupleKey = self.openfga_sdk.models.tuple_key.TupleKey

        # Make the API call
        options: Dict[str, Any] = {}
        body = ReadRequestTupleKey(user=request_user,
                                   relation=relation,
                                   object=request_object)

        # No async with here, as that would close the client
        response: ReadResponse = await self.fga_client.read(body, options)

        # Process the response
        retval: List[str] = []
        for one_tuple in response.tuples:
            # one_tuple should be a openfga_sdk.models.tuple.Tuple
            key: TupleKey = one_tuple.key

            if request_user is None:
                # We were looking for a list of actors/users.
                user: str = key.user
                user = user.split(":")[1]
                retval.append(user)

            elif request_object is None:
                # We were looking for a list of resources.
                resource: str = key.object
                resource = resource.split(":")[1]
                retval.append(resource)

            elif relation is None:
                # We were looking for a list of relations.
                retval.append(key.relation)

        return retval

    async def grant(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> bool:
        """
        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      will be permitted.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user will be permitted.

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: True if the grant succeeded, False if the grant already existed.
        """
        actor_type: str = actor.get("type", "")
        actor_id: str = actor.get("id", "")
        resource_type: str = resource.get("type", "")
        resource_id: str = resource.get("id", "")

        # Use classes from the lazily imported module to avoid extra required dependencies
        # pylint: disable=invalid-name
        ClientTuple = self.openfga_sdk.client.models.tuple.ClientTuple
        ClientWriteRequest = self.openfga_sdk.client.models.write_request.ClientWriteRequest

        client_tuple = ClientTuple(user=f"{actor_type}:{actor_id}",
                                   relation=relation,
                                   object=f"{resource_type}:{resource_id}")

        if self.debug:
            self.logger.info("Granting to %s:%s : %s on %s:%s", actor_type, actor_id,
                             relation, resource_type, resource_id)

        writes: List[ClientTuple] = []
        writes.append(client_tuple)

        await self.handle_special_user_writes(writes, actor, resource)

        body = ClientWriteRequest(writes=writes, deletes=None)

        # Hard-won tip:
        # When this call fails with a ValidationError exception (an http 400 - Bad Request)
        # what is most likely happening is either:
        #   1)  something in the write() request given does not match what is
        #       allowed to be specified in the *.fga DSL file.
        #   2)  the auth policy you think is being uploaded to the auth server
        #       in open_fga_init() is not the one actually landing in the server.
        retval: bool = True
        try:
            # No async with here, as that would close the client
            _ = await self.fga_client.write(body)

        except self.openfga_sdk.exceptions.ValidationException as err:
            if (str(err).find("tuple to be written already existed") > 0) and self.debug:
                self.logger.info("Grant already exists: %s:%s : %s on %s:%s",
                                 actor_type, actor_id, relation, resource_type, resource_id)
                retval = False
            else:
                raise

        return retval

    async def revoke(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> bool:
        """
        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      will no longer be permitted.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user will no longer be permitted.

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be no longer authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: True if the revoke succeeded, False if the revoke already existed.
        """
        actor_type: str = actor.get("type", "")
        actor_id: str = actor.get("id", "")
        resource_type: str = resource.get("type", "")
        resource_id: str = resource.get("id", "")

        # Use classes from the lazily imported module to avoid extra required dependencies
        # pylint: disable=invalid-name
        ClientTuple = self.openfga_sdk.client.models.tuple.ClientTuple
        ClientWriteRequest = self.openfga_sdk.client.models.write_request.ClientWriteRequest

        client_tuple = ClientTuple(user=f"{actor_type}:{actor_id}",
                                   relation=relation,
                                   object=f"{resource_type}:{resource_id}")

        if self.debug:
            self.logger.info("Revoking from %s:%s : %s on %s:%s", actor_type, actor_id,
                             relation, resource_type, resource_id)

        deletes: List[ClientTuple] = []
        deletes.append(client_tuple)

        body = ClientWriteRequest(writes=None, deletes=deletes)

        # Hard-won tip:
        # When this call fails with a ValidationError exception (an http 400 - Bad Request)
        # what is most likely happening is either:
        #   1)  something in the write() request given does not match what is
        #       allowed to be specified in the *.fga DSL file.
        #   2)  the auth policy you think is being uploaded to the auth server
        #       in open_fga_init() is not the one actually landing in the server.
        retval: bool = True
        try:
            # No async with here, as that would close the client
            _ = await self.fga_client.write(body)

        except self.openfga_sdk.exceptions.ValidationException as err:
            if (str(err).find("tuple to be deleted did not exist") > 0) and self.debug:
                self.logger.info("Revoke already performed: %s:%s : %s on %s:%s",
                                 actor_type, actor_id, relation, resource_type, resource_id)
                retval = False
            else:
                raise

        return retval

    async def handle_special_user_writes(self, writes: List[Any], actor: Dict[str, Any], resource: Dict[str, Any]):
        """
        This method is called from grant() or revoke() to handle special cases where
        the actor or resource is a special user.

        :writes: A list of ClientTuples to be written
        :actor: The actor dictionary with the keys "type" and "id"
        :resource: The resource dictionary with the keys "type" and "id"
        """
        _ = writes, actor, resource




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict


class ContextTypeLlmFactory:
    """
    Interface for Factory classes creating LLM BaseLanguageModels

    Most methods take a config dictionary which consists of the following keys:

        "model_name"                The name of the model.
                                    Default if not specified is "gpt-3.5-turbo"

        "temperature"               A float "temperature" value with which to
                                    initialize the chat model.  In general,
                                    higher temperatures yield more random results.
                                    Default if not specified is 0.7

        "prompt_token_fraction"     The fraction of total tokens (not necessarily words
                                    or letters) to use for a prompt. Each model_name
                                    has a documented number of max_tokens it can handle
                                    which is a total count of message + response tokens
                                    which goes into the calculation involved in
                                    get_max_prompt_tokens().
                                    By default the value is 0.5.

        "max_tokens"                The maximum number of tokens to use in
                                    get_max_prompt_tokens(). By default this comes from
                                    the model description in this class.
    """

    def load(self):
        """
        Goes through the process of loading any user extensions and/or configuration
        files
        """
        raise NotImplementedError

    def create_llm(self, config: Dict[str, Any]) -> Any:
        """
        Create an llm instance BaseLanguageModel from the fully-specified llm config.
        :param config: The fully specified llm config from which the LLM instance
                    should be created.
        :return: An llm instance native to the context type.
                Can raise a ValueError if the config's class or model_name value is
                unknown to this method.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List

from http import HTTPStatus

import json
import os
import asyncio

import tornado
from tornado.web import RequestHandler

from leaf_common.utils.async_atomic_counter import AsyncAtomicCounter
from neuro_san.internals.reservations.agent_reservation import AgentReservation
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.internals.network_providers.expiring_agent_network_storage import ExpiringAgentNetworkStorage
from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.generic.async_agent_service_provider import AsyncAgentServiceProvider
from neuro_san.service.interfaces.agent_authorizer import AgentAuthorizer
from neuro_san.service.utils.server_context import ServerContext
from neuro_san.service.http.logging.http_logger import HttpLogger


class BaseRequestHandler(RequestHandler):
    """
    Abstract handler class for neuro-san API calls.
    Provides logic to inject neuro-san service specific data
    into local handler context.
    """
    request_id_counter: AsyncAtomicCounter = AsyncAtomicCounter()

    # pylint: disable=attribute-defined-outside-init
    def initialize(self, **kwargs):
        """
        This method is called by Tornado framework to allow
        injecting service-specific data into local handler context.
        :param kwargs: dictionary of named parameters, including:
            "agent_policy" - abstract policy for agent requests;
            "forwarded_request_metadata" - list of request metadata keys to forward;
            "openapi_service_spec" - OpenAPI service spec dictionary;
            "server_context" - ServerContext instance for this server.
            "logging_config" - logging configuration dictionary
        """
        # Set up local members from kwargs dictionary passed in:
        self.agent_policy: AgentAuthorizer = kwargs.pop("agent_policy", None)
        self.forwarded_request_metadata: List[str] = kwargs.pop("forwarded_request_metadata", [])
        self.openapi_service_spec: Dict[str, Any] = kwargs.pop("openapi_service_spec", None)
        self.server_context: ServerContext = kwargs.pop("server_context", None)
        logging_config: Dict[str, Any] = kwargs.pop("logging_config", None)

        self.logger = HttpLogger(self.forwarded_request_metadata, logging_config)
        self.show_absent: bool = os.environ.get("SHOW_ABSENT_METADATA") is not None
        self.request_id: int = 0

        if os.environ.get("AGENT_ALLOW_CORS_HEADERS") is not None:
            self.set_header("Access-Control-Allow-Origin", "*")
            self.set_header("Access-Control-Allow-Methods", "GET, POST, OPTIONS")
            headers: str = "Content-Type, Transfer-Encoding"
            metadata_headers: str = ", ".join(self.forwarded_request_metadata)
            if len(metadata_headers) > 0:
                headers += f", {metadata_headers}"
            # Set all allowed headers:
            self.set_header("Access-Control-Allow-Headers", headers)

    def get_metadata(self) -> Dict[str, Any]:
        """
        Extract user metadata defined by self.forwarded_request_metadata list
        from incoming request.
        :return: dictionary of user request metadata; possibly empty
        """
        metadata_dict: Dict[str, Any] =\
            BaseRequestHandler.get_request_metadata(
                self.request,
                self.forwarded_request_metadata,
                self.show_absent,
                self.logger)
        # Put in our own unique id so we have some way to track this request:
        user_request_id: str = metadata_dict.get("request_id", "None")
        if user_request_id == "None":
            metadata_dict["request_id"] = f"request-{self.request_id}"
        return metadata_dict

    @classmethod
    def get_request_metadata(cls, request,
                             forwarded_request_metadata: List[str],
                             show_absent: bool = False,
                             logger: HttpLogger = None
                             ) -> Dict[str, Any]:
        """
        Extract user metadata defined by forwarded_request_metadata list
        from incoming request.
        :param request: incoming http request
        :param forwarded_request_metadata: list of metadata keys
        :param show_absent: if True, will provide debug printout
               for all request metadata keys absent from incoming request headers;
               if False does nothing.
        :param logger: logger to use
        :return: dictionary of user request metadata; possibly empty
        """
        headers: Dict[str, Any] = request.headers
        result: Dict[str, Any] = {}
        for item_name in forwarded_request_metadata:
            if item_name in headers.keys():
                result[item_name] = headers[item_name]
            else:
                if show_absent and logger:
                    logger.warning({}, "MISSING METADATA VALUE: %s request %s", item_name, request.uri)
                result[item_name] = "None"
        return result

    async def get_service(self, agent_name: str, metadata: Dict[str, Any]) -> AsyncAgentService:
        """
        Get agent's AsyncAgentService for request execution
        :param agent_name: agent name
        :param metadata: metadata to be used for logging if necessary.
        :return: instance of AsyncAgentService if it is defined for this agent
                 None otherwise
        """
        is_authorized: bool = False
        service_provider: AsyncAgentServiceProvider = None
        is_authorized, service_provider = await self.agent_policy.allow_agent(agent_name, metadata)

        if not is_authorized:
            # Forbidden
            self.set_status(403)
            self.do_finish()
            return None

        if service_provider is not None:
            return service_provider.get_service()

        # Now, this could be a temporary network created outside of our service:
        # Does it look like a temporary network name?
        if AgentReservation.is_reservation_name(agent_name):
            # Get temporary network storage:
            network_storage_dict: Dict[str, AgentNetworkStorage] = self.server_context.get_network_storage_dict()
            temp_storage: AgentNetworkStorage = network_storage_dict.get("temp", None)
            if temp_storage is None or not isinstance(temp_storage, ExpiringAgentNetworkStorage):
                self.set_status(500)
                self.logger.error(metadata, "error: Temporary network storage is not properly configured.")
                self.do_finish()
                return None
            # This call will trigger temp network lookup in external storage
            # and if found, it will be registered in our service
            # and become available for future requests until it expires.
            # Note that "network_provider" result is not used directly here,
            # but the call is necessary to trigger the "side effect" of lookup and registration of temp network.
            # We delegate call execution to a thread to avoid blocking server event loop
            # with potentially slow external storage access.
            network_provider = await asyncio.to_thread(temp_storage.get_agent_network_provider, agent_name)
            if network_provider is not None:
                # Now that our service is aware of this temp network,
                # we can query our policy again,
                # and we already know that this agent has been authorized:
                _, service_provider = await self.agent_policy.allow_agent(agent_name, metadata)
                # If service provider is still None,
                # that means something went very wrong - bail out:
                if service_provider is not None:
                    return service_provider.get_service()

        # Nothing worked - so we report resource not found
        self.set_status(404)
        self.logger.error(metadata, "error: Invalid request path %s", self.request.path)
        self.do_finish()
        return None

    def process_exception(self, exc: Exception):
        """
        Process exception raised during request handling
        """
        if exc is None:
            return
        if isinstance(exc, json.JSONDecodeError):
            # Handle invalid JSON input
            self.set_status(HTTPStatus.BAD_REQUEST)
            self.write({"error": "Invalid JSON format"})
            self.logger.error(self.get_metadata(), "error: Invalid JSON format")
            return

        # General exception case:
        self.set_status(HTTPStatus.INTERNAL_SERVER_ERROR)
        self.write({"error": "Internal server error"})
        self.logger.error(self.get_metadata(), "Internal server error: %s", str(exc))

    def data_received(self, chunk):
        """
        Method overrides abstract method of RequestHandler
        with no-op implementation.
        """
        return

    # Tornado can handle both syns and async versions of "prepare" method
    # pylint: disable=invalid-overridden-method
    async def prepare(self):
        if not self.application.is_serving():
            self.set_status(503)
            self.write({"error": "Server is shutting down"})
            self.logger.error(self.get_metadata(), "Server is shutting down")
            self.do_finish()
            return

        # Get unique request id in case we will need it:
        self.request_id = await self.request_id_counter.increment()

        self.logger.debug(self.get_metadata(), f"[REQUEST RECEIVED] {self.request.method} {self.request.uri}")

    def do_finish(self):
        """
        Wrapper for finish() call
        with check for closed client connection.
        """
        try:
            self.finish()
        except tornado.iostream.StreamClosedError:
            self.logger.warning(self.get_metadata(), "Finish: client closed connection unexpectedly.")

    async def do_flush(self) -> bool:
        """
        Wrapper for flush() call
        with check for closed client connection.
        """
        try:
            await self.flush()
            # What happens here: we have finished writing out one data item in our output stream,
            # and we have flushed Tornado output.
            # BUT: this does not guarantee in general that underlying TCP/IP transport
            # will flush its own buffers, so low-level buffering is still possible.
            # Result would be that several chat responses will be bunched together
            # and received by a client as one data piece.
            # If client is not ready for this, there will be problems.
            # SO: this real wall clock delay here helps to encourage underlying transport
            # to flush its own buffers - and we are good.
            # Duration of delay is speculative and maybe could be adjusted.
            # But best solution and reliable one: make client accept multiple data items
            # in one "get" request - as it should when dealing with streaming service.
            await asyncio.sleep(0.3)
            return True
        except tornado.iostream.StreamClosedError:
            self.logger.warning(self.get_metadata(), "Flush: client closed connection unexpectedly.")
            return False

    async def options(self, *_args, **_kwargs):
        """
        Handles OPTIONS requests for CORS support
        """
        # No body needed. Tornado will return a 204 No Content by default
        self.set_status(HTTPStatus.NO_CONTENT)
        self.do_finish()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.filters.message_filter import MessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType
from neuro_san.message_processing.message_processor import MessageProcessor


class CompositeMessageProcessor(MessageProcessor):
    """
    A MessageProcessor implementation that employs multiple MessageProcessors.
    """

    def __init__(self, message_processors: List[MessageProcessor] = None,
                 depth_blocks_breadth: bool = False):
        """
        Constructor

        :param message_processors: An ordered List of MessageProcessors with which
                     this instance will process messages
        :param depth_blocks_breadth: When True, any one component of this instance (depth)
                    can block further message processing of downstream peers (breadth).
        """
        self.message_processors: List[MessageProcessor] = message_processors
        if self.message_processors is None:
            self.message_processors = []
        self.depth_blocks_breadth: bool = depth_blocks_breadth

    def add_processor(self, message_processor: MessageProcessor):
        """
        Adds a MessageProcessor to the list.
        Order can be important.

        :param message_processor: The MessageProcessor to add
        """
        if message_processor is not None:
            self.message_processors.append(message_processor)

    def reset(self):
        """
        Resets any previously accumulated state
        """
        for message_processor in self.message_processors:
            message_processor.reset()

    def should_block_downstream_processing(self, chat_message_dict: Dict[str, Any],
                                           message_type: ChatMessageType) -> bool:
        """
        :param chat_message_dict: The ChatMessage dictionary to consider.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        :return: True if the given message should be blocked from further downstream
                processing.  False otherwise (the default).
        """
        if not self.depth_blocks_breadth:
            return False

        for message_processor in self.message_processors:
            if message_processor.should_block_downstream_processing(chat_message_dict, message_type):
                return True
        return False

    def process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType = None):
        """
        Process the message.
        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
                            Can be None to kick the process off.
        """
        if message_type is None:
            message_type = MessageFilter.get_message_type(chat_message_dict)

        for message_processor in self.message_processors:
            message_processor.process_message(chat_message_dict, message_type)
            if message_processor.should_block_downstream_processing(chat_message_dict, message_type):
                break

    async def async_process_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType = None):
        """
        Process the message asynchronously.

        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
                            Can be None to kick the process off.
        """
        if message_type is None:
            message_type = MessageFilter.get_message_type(chat_message_dict)

        for message_processor in self.message_processors:
            await message_processor.async_process_message(chat_message_dict, message_type)
            if message_processor.should_block_downstream_processing(chat_message_dict, message_type):
                break




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from leaf_common.time.timeout import Timeout
from leaf_common.asyncio.asyncio_executor_pool import AsyncioExecutorPool

from neuro_san.client.direct_agent_storage_util import DirectAgentStorageUtil
from neuro_san.interfaces.agent_session import AgentSession
from neuro_san.internals.interfaces.context_type_toolbox_factory import ContextTypeToolboxFactory
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.run_context.factory.master_toolbox_factory import MasterToolboxFactory
from neuro_san.internals.run_context.factory.master_llm_factory import MasterLlmFactory
from neuro_san.internals.graph.persistence.agent_network_restorer import AgentNetworkRestorer
from neuro_san.internals.graph.persistence.registry_manifest_restorer import RegistryManifestRestorer
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.internals.network_providers.expiring_agent_network_storage import ExpiringAgentNetworkStorage
from neuro_san.internals.reservations.direct_agent_reservationist import DirectAgentReservationist
from neuro_san.session.direct_agent_session import DirectAgentSession
from neuro_san.session.external_agent_session_factory import ExternalAgentSessionFactory
from neuro_san.session.missing_agent_check import MissingAgentCheck
from neuro_san.session.session_invocation_context import SessionInvocationContext


class DirectAgentSessionFactory:
    """
    Sets up everything needed to use a DirectAgentSession more as a library.
    This includes:
        * Some reading of AgentNetworks
        * Setting up AgentNetworkStorage with agent networks
          which were read in
        * Initializing an LlmFactory
    """

    def __init__(self):
        """
        Constructor
        """
        # Read the manifest once and pass that into the Util call below.
        manifest_restorer = RegistryManifestRestorer()
        manifest_networks: Dict[str, Dict[str, AgentNetwork]] = manifest_restorer.restore()

        self.network_storage_dict: Dict[str, AgentNetworkStorage] = {
            "temp": ExpiringAgentNetworkStorage()
        }

        for storage_type in ["public", "protected"]:
            storage: AgentNetworkStorage = DirectAgentStorageUtil.create_network_storage(manifest_networks,
                                                                                         storage_type=storage_type)
            self.network_storage_dict[storage_type] = storage

    def create_session(self, agent_name: str, use_direct: bool = False,
                       metadata: Dict[str, str] = None, umbrella_timeout: Timeout = None) -> AgentSession:
        """
        :param agent_name: The name of the agent to use for the session.
                This name can be something in the manifest file (with no file suffix)
                or a specific full-reference to an agent network's hocon file.
        :param use_direct: When True, will use a Direct session for
                    external agents that would reside on the same server.
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param umbrella_timeout: A Timeout object to periodically check in loops.
                        Default is None (no timeout).
        """

        agent_network: AgentNetwork = self.get_agent_network(agent_name)
        config: Dict[str, Any] = agent_network.get_config()
        llm_factory: ContextTypeLlmFactory = MasterLlmFactory.create_llm_factory(config)
        toolbox_factory: ContextTypeToolboxFactory = MasterToolboxFactory.create_toolbox_factory(config)
        # Load once now that we know what tool registry to use.
        llm_factory.load()
        toolbox_factory.load()

        factory = ExternalAgentSessionFactory(use_direct=use_direct, network_storage_dict=self.network_storage_dict)
        executors_pool = AsyncioExecutorPool()

        # DEF - We could do max_lifetime here, but waiting until that seems necessary.
        reservationist = DirectAgentReservationist(set([self.network_storage_dict.get("temp")]))
        invocation_context = SessionInvocationContext(agent_name,
                                                      factory,
                                                      executors_pool,
                                                      llm_factory,
                                                      toolbox_factory,
                                                      metadata,
                                                      reservationist)
        invocation_context.start()
        session: DirectAgentSession = DirectAgentSession(agent_network=agent_network,
                                                         invocation_context=invocation_context,
                                                         metadata=metadata,
                                                         umbrella_timeout=umbrella_timeout)
        return session

    def get_agent_network(self, agent_name: str) -> AgentNetwork:
        """
        :param agent_name: The name of the agent whose AgentNetwork we want to get.
                This name can be something in the manifest file (with no file suffix)
                or a specific full-reference to an agent network's hocon file.
        :return: The AgentNetwork corresponding to that agent.
        """

        if agent_name is None or len(agent_name) == 0:
            return None

        agent_network: AgentNetwork = None
        if agent_name.endswith(".hocon") or agent_name.endswith(".json"):
            # We got a specific file name
            restorer = AgentNetworkRestorer()
            agent_network = restorer.restore(file_reference=agent_name)
        else:
            # Use the standard stuff available via the manifest file.
            for storage_type in ["public", "protected"]:
                storage: AgentNetworkStorage = self.network_storage_dict.get(storage_type)
                agent_network_provider: AgentNetworkProvider = storage.get_agent_network_provider(agent_name)
                if agent_network_provider is None:
                    continue
                agent_network = agent_network_provider.get_agent_network()
                if agent_network is not None:
                    break

        # Common place for nice error messages when networks are not found
        MissingAgentCheck.check_agent_network(agent_network, agent_name)

        return agent_network




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from copy import deepcopy
from threading import Lock

from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.service.interfaces.event_loop_logger import EventLoopLogger
from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.generic.agent_server_logging import AgentServerLogging
from neuro_san.service.utils.server_context import ServerContext


# pylint: disable=too-many-instance-attributes
class AsyncAgentServiceProvider:
    """
    Class providing lazy construction of AsyncAgentService instance
    with given constructor parameters.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self,
                 request_logger: EventLoopLogger,
                 security_cfg: Dict[str, Any],
                 agent_name: str,
                 agent_network_provider: AgentNetworkProvider,
                 server_logging: AgentServerLogging,
                 server_context: ServerContext):
        """
        Constructor.
        :param request_logger: The instance of the EventLoopLogger that helps
                        log information from running event loop
        :param security_cfg: A dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  If None, uses insecure channel.
        :param agent_name: The agent name for the service
        :param agent_network_provider: The AgentNetworkProvider to use for the session.
        :param server_logging: An AgentServerLogging instance initialized so that
                        spawned asynchronous threads can also properly initialize
                        their logging.
        :param server_context: The ServerContext object containing global-ish state
        """
        self.request_logger = request_logger
        self.security_cfg = deepcopy(security_cfg)
        self.server_logging: AgentServerLogging = server_logging
        self.agent_network_provider: AgentNetworkProvider = agent_network_provider
        self.agent_name: str = agent_name
        self.lock: Lock = Lock()
        self.server_context: ServerContext = server_context
        self.service_instance: AsyncAgentService = None

    def get_service(self) -> AsyncAgentService:
        """
        Get service instance.
        Create it if it is not instantiated yet.
        """
        if self.service_instance is None:
            with self.lock:
                if self.service_instance is None:
                    self.service_instance = AsyncAgentService(
                        self.request_logger,
                        self.security_cfg,
                        self.agent_name,
                        self.agent_network_provider,
                        self.server_logging,
                        self.server_context)
        return self.service_instance

    def reset_service(self):
        """
        Reset service instance to None to allow re-creation on next get_service() call
        in case underlying AgentNetwork has changed.
        """
        with self.lock:
            self.service_instance = None

    def service_created(self) -> bool:
        """
        Return True if service instance has already been instantiated;
               False otherwise
        """
        return self.service_instance is not None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import Union


from logging import getLogger
from logging import Logger

from neuro_san.interfaces.coded_tool import CodedTool


class Accountant(CodedTool):
    """
    A tool that updates a running cost each time it is called.
    """

    async def async_invoke(self, args: Dict[str, Any], sly_data: Dict[str, Any]) -> Union[Dict[str, Any], str]:
        """
        Updates the passed running cost each time it's called.
        :param args: A dictionary with optional key:
                    "running_cost": the running cost to update (optional if in sly_data).

        :param sly_data: A dictionary containing parameters that should be kept out of the chat stream.
                         Keys expected for this implementation are:
                         "running_cost": the running cost to update (optional if in args).

        :return: A dictionary containing:
                 "running_cost": the updated running cost.
                 Also updates sly_data with the new running cost if it was the source.
        """
        tool_name = self.__class__.__name__
        logger: Logger = getLogger(self.__class__.__name__)

        logger.debug("========== Calling %s ==========", tool_name)
        logger.debug("args: %s", str(args))

        # Try to get running_cost from args first, then sly_data, then default to 0.0
        if "running_cost" in args:
            running_cost: float = float(args.get("running_cost"))
        else:
            running_cost: float = float(sly_data.get("running_cost", 0.0))

        # Increment the running cost not using value other than 1
        # This would make it a little harder if the LLM wanted to guess
        updated_running_cost: float = running_cost + 3.0

        # Update sly_data if running_cost came from it
        if "running_cost" not in args:
            sly_data["running_cost"] = updated_running_cost

        tool_response = {
            "running_cost": updated_running_cost
        }
        logger.debug("-----------------------")
        logger.debug("%s response: %s", tool_name, tool_response)
        logger.debug("========== Done with %s ==========", tool_name)
        return tool_response




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from os import environ

from asyncio import AbstractEventLoop
from asyncio import Event
from asyncio import run_coroutine_threadsafe
from asyncio import sleep as asyncio_sleep
from asyncio import CancelledError
from logging import getLogger
from logging import Logger
import queue as queue_mod

from janus import Queue
from janus import SyncQueueShutDown

from leaf_common.config.resolver_util import ResolverUtil
from leaf_common.asyncio.asyncio_executor import AsyncioExecutor

from neuro_san.internals.chat.async_collating_queue import AsyncCollatingQueue
from neuro_san.internals.interfaces.reservations_storage import ReservationsStorage
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.internals.network_providers.expiring_agent_network_storage import ExpiringAgentNetworkStorage
from neuro_san.internals.reservations.agent_reservation import AgentReservation
from neuro_san.internals.reservations.abstract_agent_reservationist import AbstractAgentReservationist
from neuro_san.internals.interfaces.startable import Startable


class TempNetworkStorageUpdater(Startable):
    """
    StorageUpdater implementation for temporary network being generated by neuro-san service.
    """

    def __init__(self, network_storage_dict: Dict[str, AgentNetworkStorage],
                 queues: Queue[AsyncCollatingQueue]):
        """
        Constructor

        :param network_storage_dict: A dictionary of string (descripting scope) to
                    AgentNetworkStorage instance which keeps all the AgentNetwork instances
                    of a particular grouping.
        :param queues: A Queue of AsyncCollatingQueues for temp network deployment
        """
        self.logger: Logger = getLogger(self.__class__.__name__)

        self.reservationist: AbstractAgentReservationist = None
        self.executor: AsyncioExecutor = None
        self.temp_storage: ReservationsStorage = network_storage_dict.get("temp")
        if self.temp_storage is not None:
            # If we don't have temp storage, we don't got nothin'
            # Check that the temp storage instance is indeed an ExpiringReservationsStorage,
            # since we need to be able to expire reservations in it.
            if not isinstance(self.temp_storage, ExpiringAgentNetworkStorage):
                raise ValueError("Temp reservations storage must be an instance of ExpiringAgentNetworkStorage")

            # Potentially create a "base" reservations storage class
            storage_class_name: str = environ.get("AGENT_EXTERNAL_RESERVATIONS_STORAGE", "")
            external_storage: ReservationsStorage = ResolverUtil.create_instance(
                    storage_class_name,
                    "AGENT_EXTERNAL_RESERVATIONS_STORAGE env var",
                    ReservationsStorage)
            if external_storage is not None:
                self.temp_storage.set_base_storage(external_storage)
            self.reservationist = AbstractAgentReservationist({self.temp_storage})

            # We use an AsyncioExecutor and its event loop to process the collection of tasks
            # this TempNetworkStorageUpdater instance is responsible for:
            # 1. Checking for new queues to process coming in self.incoming "main" queue
            # 2. Processing the individual queues that come in,
            #    which involves processing the items that come in on those queues
            #    and deploying reservations to the temp storage, which is done by self.reservationist.
            self.executor = AsyncioExecutor()

            self.incoming: Queue[AsyncCollatingQueue] = queues
        else:
            # We don't have a temp storage, so we won't be doing any processing,
            # close the incoming queues and don't start any processing threads.
            self.logger.info("No temp storage configured, closing incoming queues without processing")
            queues.close()

    def start(self):
        """
        Perform start up.
        """
        if self.temp_storage is None:
            self.logger.info("TempNetworkStorageUpdater is not started: no temp storage configured.")
            return

        self.logger.info("Starting TempNetworkStorageUpdater")
        # Start any Startables
        if isinstance(self.temp_storage, Startable):
            self.temp_storage.start()

        if self.executor is not None:
            self.executor.start()
            # Start the task for checking for new queues to process,
            # this task is expected to run for the lifetime of this TempNetworkStorageUpdater instance.
            _ = self.executor.create_task(self.add_new_queues_to_processing(), "incoming queue")

    async def add_new_queues_to_processing(self):
        """
        Checks our master queue of queues for any new additions
        and adds processing tasks to executor
        for independent processing of each queue.
        """
        idle_sleep_seconds: float = 2.0

        self.logger.info("Waiting for new queues to be added for processing")
        while True:
            try:
                async_collating_queue: AsyncCollatingQueue = self.incoming.sync_q.get_nowait()
            except queue_mod.Empty:
                # No item to process right now, sleep for a bit and check again.
                # But first check, if the queue is closed already:
                if self.incoming.sync_q.closed:
                    self.logger.info("Incoming queue closed")
                    return
                try:
                    await asyncio_sleep(idle_sleep_seconds)
                except CancelledError:
                    self.logger.info("Incoming queue task cancelled during sleep")
                    raise
                # Continue polling for new queues to process.
                continue
            except SyncQueueShutDown:
                # Explicit shutdown signal from Janus queue machinery
                self.logger.info("SHUTDOWN signal for incoming queue")
                return

            if self.reservationist is not None:
                _ = self.executor.create_task(self.process_one_queue(async_collating_queue), "queue_processor")
            else:
                # We don't have a reservation functionality set up,
                # so we can't do anything with this queue, but we should still close it to avoid leaks.
                self.logger.warning("No reservationist available, closing incoming queue without processing")
                async_collating_queue.close()

    async def process_one_queue(self, async_collating_queue: AsyncCollatingQueue):
        """
        Processes a single AsyncCollatingQueue

        :param async_collating_queue: The AsyncCollatingQueue to process
        """
        janus_queue: Queue = async_collating_queue.get_queue()
        idle_sleep_seconds: float = 0.5

        # See what has come over this particular queue
        while True:
            # Get an item off the queue
            try:
                queued_item: Dict[str, Any] = janus_queue.sync_q.get_nowait()
            except queue_mod.Empty:
                # No item to process right now, sleep for a bit and check again.
                # But first check, if the queue is closed already:
                if janus_queue.sync_q.closed:
                    return
                try:
                    await asyncio_sleep(idle_sleep_seconds)
                except CancelledError:
                    self.logger.info("Reservation queue task cancelled during sleep")
                    raise
                continue
            except SyncQueueShutDown:
                # Explicit shutdown signal from Janus queue machinery
                self.logger.info("SHUTDOWN signal for reservation queue")
                return
            if async_collating_queue.is_final_item(queued_item):
                # We have exhausted this queue.
                # No one needs to worry about it anymore.
                async_collating_queue.close()
                return

            # Now: we have an item to process from this queue.
            self.process_one_queued_item(queued_item)

    def process_one_queued_item(self, queued_item: Dict[str, Any]):
        """
        Process a single item from one of the queues

        :param queued_item: A dictionary from a queue containing information
                as populated by AgentReservationist.deploy()
        """

        # Get the salient information from the dictionary that was queued.
        source: str = queued_item.get("source")
        deployment_dict: Dict[AgentReservation, Dict[str, Any]] = queued_item.get("deployment_dict")
        max_lifetime_in_seconds: float = queued_item.get("max_lifetime_in_seconds")

        # Do the deployment
        self.reservationist.deploy_together(deployment_dict, source, max_lifetime_in_seconds)

        # Maybe notify the deployer.
        event: Event = queued_item.get("event")
        if event is not None:
            event_loop: AbstractEventLoop = queued_item.get("event_loop")
            run_coroutine_threadsafe(self.reservationist.set_event(event), event_loop)

    def stop(self):
        """
        Perform stopping.
        """
        self.logger.info("Stopping TempNetworkStorageUpdater")
        if self.temp_storage is not None:
            self.temp_storage.stop()

        # Finalize the executor:
        if self.executor is not None:
            self.executor.shutdown()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple

import logging
import time

from neuro_san.interfaces.reservation import Reservation
from neuro_san.internals.graph.registry.agent_network import AgentNetwork
from neuro_san.internals.interfaces.agent_network_provider import AgentNetworkProvider
from neuro_san.internals.interfaces.reservations_storage import ReservationsStorage
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.internals.network_providers.abstract_reservations_storage \
    import AbstractReservationsStorage
from neuro_san.internals.network_providers.fixed_agent_network_provider import FixedAgentNetworkProvider
from neuro_san.internals.interfaces.startable import Startable


class ExpiringAgentNetworkStorage(AbstractReservationsStorage, AgentNetworkStorage):
    """
    An AgentNetworkStorage instance where AgentNetworks are allowed to expire.
    This implementation also allows for an optional "source" storage to be set,
    which will be used as a "base" source of truth.
    """

    def __init__(self, check_expirations_interval_seconds: float = 60.0):
        """
        Constructor
        :param check_expirations_interval_seconds: The number of seconds between checks for expired reservations.
        """
        super().__init__(storage_name="temp_storage",
                         check_expirations_interval_seconds=check_expirations_interval_seconds)
        self.reservations_table: Dict[str, Reservation] = {}
        self.base_storage: ReservationsStorage = None
        self.logger = logging.getLogger(self.__class__.__name__)

    def set_base_storage(self, base_storage: ReservationsStorage):
        """
        Set the base storage to be used as a "source of truth" for reservations and agent networks.
        This is optional, but if set, then this storage will be used as a source of truth
        for reservations and agent networks,
        and any changes to reservations and agent networks in this storage will be reflected in this storage as well.

        :param base_storage: An ExpiringReservationsStorage instance to be used
                             as a source of truth for reservations and agent networks.
        """
        self.base_storage = base_storage

    def start(self):
        """
        Start this storage, which includes starting the expiration checking loop.
        """
        super().start()
        try:
            if self.base_storage is not None:
                self.base_storage.start()
        except Exception as exc:  # pylint: disable=broad-except
            self.logger.error("%s: Failed to start base storage: %s", self._name, exc)
            super().stop()
            raise

    def stop(self, timeout: Optional[float] = None):
        """
        Stop this storage, which includes stopping the expiration checking loop.
        """
        super().stop()
        try:
            if self.base_storage is not None and isinstance(self.base_storage, Startable):
                self.base_storage.stop()
        except Exception as exc:  # pylint: disable=broad-except
            self.logger.error("%s: Failed to stop base storage: %s", self._name, exc)
            raise

    @staticmethod
    def is_expired(reservation: Reservation) -> bool:
        """
        :return: Whether this reservation is expired or not.
        """
        return time.time() > reservation.get_expiration_time_in_seconds()

    def add_reservations(self, reservations_dict: Dict[Reservation, Dict[str, Any]],
                         source: str = None):
        """
        Add a set of reservations for agent networks en-masse

        :param reservations_dict: A mapping of Reservation -> agent network spec
        :param source: A string describing where the deployment was coming from
        """
        if not reservations_dict:
            # Nothing to do
            return

        if self.base_storage is not None:
            # If we have a source storage, then we need to add these reservations there first:
            self.base_storage.add_reservations(reservations_dict, source=source)

        # Figure out what's new vs what's not.
        # Need to do this while holding the lock
        added: List[str] = []
        replaced: List[str] = []
        with self.lock:
            for reservation, agent_spec in reservations_dict.items():

                agent_name: str = reservation.get_reservation_id()
                is_new = self.agents_table.get(agent_name) is None

                agent_network = AgentNetwork(agent_spec, agent_name)
                self.agents_table[agent_name] = agent_network
                self.reservations_table[agent_name] = reservation

                if is_new:
                    added.append(agent_name)
                else:
                    replaced.append(agent_name)

        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            for agent_name in added:
                listener.agent_added(agent_name, self)
                self.logger.info("%s: ADDED network for agent %s from %s", self._name, agent_name, source)
            for agent_name in replaced:
                listener.agent_modified(agent_name, self)
                self.logger.info("%s: REPLACED network for agent %s from %s", self._name, agent_name, source)

    def get_one_reservation(self, obj_key: str) -> Tuple[Reservation, Any]:
        """
        Extract a single reservation.

        :param obj_key: unique key for the reservation
        :return: Tuple of (reservation, agent data) if successful
                 and reservation is not expired,
                 (None, None) otherwise
        """
        # For this class, this method is never used.
        raise NotImplementedError

    def expire_reservations(self):
        """
        Remove Reservations that are expired
        """

        # First determine what has expired
        expired: List[str] = []

        reservations_snapshot: Dict[str, Reservation] = None
        with self.lock:
            reservations_snapshot = self.reservations_table.copy()

        for agent_name, reservation in reservations_snapshot.items():
            if self.is_expired(reservation):
                expired.append(agent_name)

        # Nothing to do?
        if len(expired) == 0:
            return

        # Do the dirty deeds.
        with self.lock:
            for agent_name in expired:
                self.reservations_table.pop(agent_name, None)
                self.agents_table.pop(agent_name, None)

        # Notify listeners about this state change:
        # do it outside of internal lock
        for listener in self.listeners:
            for agent_name in expired:
                listener.agent_removed(agent_name, self)
                self.logger.info("%s: REMOVED network for agent %s", self._name, agent_name)

    def get_agent_network_provider(self, agent_name: str) -> AgentNetworkProvider:
        """
        Get AgentNetworkProvider for a specific agent
        :param agent_name: name of an agent
        """
        # Agents and their Reservations are always present or absent together.
        reservation: Reservation = self.reservations_table.get(agent_name, None)
        if reservation is not None:
            # We have a reservation for this agent, but we need to check is it still valid:
            if self.is_expired(reservation):
                # Reservation is expired, so AgentNetwork is gone for good:
                with self.lock:
                    self.reservations_table.pop(agent_name, None)
                    self.agents_table.pop(agent_name, None)

                # Notify listeners about this state change:
                # do it outside of internal lock
                for listener in self.listeners:
                    listener.agent_removed(agent_name, self)
                    self.logger.info("%s: REMOVED expired network for agent %s", self._name, agent_name)

                # Now if we have a source storage,
                # we rely on it to expire this reservation according to its own schedule.

                # Our agent network no longer exists.
                return None
            # Reservation is still valid, so we can return the AgentNetworkProvider for this agent.
            agent_network: AgentNetwork = self.agents_table.get(agent_name, None)
            if agent_network is not None:
                return FixedAgentNetworkProvider(agent_network)
            return None
        # We don't have a reservation for this agent,
        # so check the source storage if we have one:
        if self.base_storage is not None:
            reservation, agent_network = self.base_storage.get_one_reservation(agent_name)
            if reservation is not None and not self.is_expired(reservation):
                # We have a valid reservation for this agent in the source storage,
                # so return the AgentNetworkProvider for this agent.
                # Also cache this reservation and agent network locally:
                with self.lock:
                    self.reservations_table[agent_name] = reservation
                    self.agents_table[agent_name] = agent_network
                # Notify listeners about this state change:
                # do it outside of internal lock
                for listener in self.listeners:
                    listener.agent_added(agent_name, self)
                    self.logger.info("ADDED network for agent %s from %s", agent_name, "base storage")

                return FixedAgentNetworkProvider(agent_network)
        return None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

class ServiceStatus:
    """
    Class for registering and reporting overall status of the service,
    primarily for interaction with external deployment environment.
    """

    def __init__(self, service_name: str):
        """
        Constructor.
        """
        self.service_name: str = service_name
        self.service_requested: bool = True
        self.service_ready: bool = False

    def set_status(self, status: bool):
        """
        Set the status of a service
        """
        self.service_ready = status

    def is_ready(self) -> bool:
        """
        True if service is ready
        """
        return self.service_ready

    def set_requested(self, requested: bool):
        """
        Set if a service is requested by neuro-san server.
        """
        self.service_requested = requested

    def is_requested(self) -> bool:
        """
        True if service is requested.
        """
        return self.service_requested

    def get_service_name(self) -> str:
        """
        Return service name
        """
        return self.service_name




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from leaf_common.parsers.dictionary_extractor import DictionaryExtractor

from neuro_san.internals.run_context.interfaces.agent_network_inspector import AgentNetworkInspector


class InvocationUtil:
    """
    Utility class for handling questions about agent network invocations.
    """

    @staticmethod
    def get_effective_invocation(inspector: AgentNetworkInspector, request_dict: Dict[str, Any]) -> str:
        """
        Check if the agent network should process the request as an event.

        :param inspector: The AgentNetworkInspector instance.  Common internal implementations of
                        AgentNetworkInspector include AgentNetwork, AgentToolRegistry.
        :param request_dict: A ChatRequest dictionary.

        :return: A string representing the invocation type.
                 Can be "event" or "chatbot" for now.
        """
        if inspector is None:
            return "chatbot"

        invocation: str = None
        front_man: str = None
        try:
            front_man = inspector.find_front_man()
        except ValueError:
            # No front man found.
            pass

        if front_man is not None:
            front_man_spec: Dict[str, Any] = inspector.get_agent_tool_spec(front_man)
            spec_extractor = DictionaryExtractor(front_man_spec)
            # Default invocation if none specified is chatbot
            invocation = spec_extractor.get("function.invocation", "chatbot")

        chat_filter: str = "MINIMAL"
        if request_dict is not None:
            request_extractor = DictionaryExtractor(request_dict)
            chat_filter = request_extractor.get("chat_filter.chat_filter_type", chat_filter)

        # More possibilities will come for more invocations and chat filters.
        if invocation in ("event",) and chat_filter in ("MINIMAL",):
            return "event"

        # Note that even if the spec wants event but the request is a MAXIMAL filter,
        # we are taking that to mean that the client wants to hang around for the queued messages
        # to debug what is going on with the event.
        return "chatbot"




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from leaf_common.time.timeout import Timeout

from neuro_san.interfaces.agent_session_constants import AgentSessionConstants


class AbstractHttpServiceAgentSession(AgentSessionConstants):
    """
    Abstract base class for sync and async version for http service agent sessions.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(self, host: str = None,
                 port: str = None,
                 timeout_in_seconds: int = 30,
                 metadata: Dict[str, str] = None,
                 security_cfg: Dict[str, Any] = None,
                 umbrella_timeout: Timeout = None,
                 streaming_timeout_in_seconds: int = None,
                 agent_name: str = None):
        """
        Creates a AsyncAgentSession that asynchronously connects to the
        Agent Service and delegates its implementations to the service.

        :param host: the service host to connect to
                        If None, will use a default
        :param port: the service port
                        If None, will use a default
        :param timeout_in_seconds: timeout to use when communicating
                        with the service
        :param metadata: A grpc metadata of key/value pairs to be inserted into
                         the header. Default is None. Preferred format is a
                         dictionary of string keys to string values.
        :param security_cfg: An optional dictionary of parameters used to
                        secure the TLS and the authentication of the gRPC
                        connection.  Supplying this implies use of a secure
                        GRPC Channel.  Default is None, uses insecure channel.
        :param umbrella_timeout: A Timeout object under which the length of all
                        looping and retries should be considered
        :param streaming_timeout_in_seconds: timeout to use when streaming to/from
                        the service. Default is None, indicating connection should
                        stay open until the (last) result is yielded.
        :param agent_name: The name of the agent to talk to
        """
        _ = umbrella_timeout

        self.security_cfg: Dict[str, Any] = security_cfg
        self.use_host: str = "localhost"
        if host is not None:
            self.use_host = host

        self.use_port: str = str(self.DEFAULT_HTTP_PORT)
        if port is not None:
            self.use_port = port

        self.agent_name: str = agent_name

        self.timeout_in_seconds: int = timeout_in_seconds
        self.streaming_timeout_in_seconds: int = streaming_timeout_in_seconds
        self.metadata: Dict[str, str] = metadata

    def get_request_path(self, method: str) -> str:
        """
        :param method: The method endpoint we wish to reach
        :return: The full URL for accessing the method, given the host, port and agent_name.
        """
        scheme: str = "http"
        if self.security_cfg is not None:
            scheme = "https"

        if self.agent_name is None:
            return f"{scheme}://{self.use_host}:{self.use_port}/api/v1/{method}"
        return f"{scheme}://{self.use_host}:{self.use_port}/api/v1/{self.agent_name}/{method}"

    def get_headers(self) -> Dict[str, Any]:
        """
        Get headers for any outgoing request
        """
        headers: Dict[str, Any] = self.metadata
        if headers is None:
            headers = {}
        return headers

    def help_message(self, path: str) -> str:
        """
        Method returning general help message for http connectivity problems.
        :param path: url path of a request
        :return: help message
        """
        message = f"""
        Some basic suggestions to help debug connectivity issues:
        1. Ensure the server is running and reachable:
           ping <server_address>
           curl -v <server_url>
        2. Check network issues:
           traceroute <server_address>  # Linux/macOS
           tracert <server_address>  # Windows
        3. Ensure you are using correct protocol (http/https) and port number;
        4. Run service health check:
           curl <server_url:server_port>
        5. Try testing with increased timeout;
        6. Did you misspell the agent and/or method name in your {path} request path?
        7. If working with a local docker container:
           7.1 Does your http port EXPOSEd in the Dockerfile match your value for AGENT_HTTP_PORT?
           7.2 Did you add a -p <server_port>:<server_port> to your docker run command line to map container port(s)
               to your local ones?
        8. Is the agent turned on in your manifest.hocon?
        9. If you are attempting to use https, know that the default server configurations do not
           provide any of the necessary certificates for this to work and any certs used will
           need to be well known.  If you're unfamiliar with this process, it's a big deal.
           Try regular http instead.
        10. Is the server your are attempting to reach running with an AGENT_AUTHORIZER?
            If so, have the administrator for the server check its authorization
            policy/database to be sure the your user has the correct permissions for the network.
        """
        return message




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.authorization.interfaces.abstract_authorizer import AbstractAuthorizer


class AlwaysNoAuthorizer(AbstractAuthorizer):
    """
    Implementation of the Authorizer interface that lets no requests through.
    """

    async def authorize(self, actor: Dict[str, Any], action: str, resource: Dict[str, Any]) -> bool:
        """
        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      is seeking permission.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param action:  The action for which the user is asking permission for.
                        Most often this is one of the Permission values of:
                            "create", "read", "update" or "delete".
        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: True if the actor is allowed to take the requested action on the resource.
                 False otherwise.
        """
        # By default, no one can do anything
        return False

    async def grant(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> bool:
        """
        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      will be permitted.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user will be permitted.
                     Most often this is one of the strings from the Role enum.

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: True if the grant succeeded, False if the grant already existed.
        """
        return False

    async def revoke(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> bool:
        """
        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      will no longer be permitted.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user will no longer be permitted.
                     Most often this is one of the strings from the Role enum.

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be no longer authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            "id": "hello_world"
                        }
        :return: True if the revoke succeeded, False if the revoke already existed.
        """
        return False

    async def list(self, actor: Dict[str, Any], relation: str, resource: Dict[str, Any]) -> List[str]:
        """
        Return a list of resource ids that the actor has the given relation to,
        as per the graph specified by the authorization model.

        :param actor: The actor dictionary with the keys "type" and "id" identifying what
                      entity's relation should be checked.  Most often this is of the form:
                        {
                            "type": "User",
                            "id": "<username>"
                        }
        :param relation: The relation for which the user's permissions will be checked.
                        Most often this is one of the Permission values of:
                            "create", "read", "update" or "delete".

        :param resource: The resource dictionary with the keys "type" and "id" identifying
                      just what is to be authorized for use.  For instance:
                        {
                            "type": "AgentNetwork",
                            # Note: "id" is not specified. We want a list of these returned.
                        }
        :return: A list of resource ids that the actor has the given relation with.
                 An empty return list implies that the actor has access to no objects
                 of the given resource type.
        """
        # Indicate no access to anything
        retval: List[str] = []
        return retval




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from enum import Enum

from leaf_common.serialization.interface.dictionary_converter import DictionaryConverter

from neuro_san.internals.messages.chat_message_type import ChatMessageType


class ChatMessageConverter(DictionaryConverter):
    """
    Helper class to prepare chat response messages
    for external clients consumption.
    """
    def to_dict(self, obj: object) -> Dict[str, object]:
        """
        :param obj: The object (chat response) to be converted into a dictionary
        :return: chat response dictionary in format expected by clients
        """
        # Do "the safe" copy, moving over only json serializable values.
        # Note that values of ChatMessageType are passed through,
        # because they are handled by a separate post-processing step.
        response_dict = self.to_json_safe(obj)
        # This is where ChatMessageType enum values are converted:
        self.convert(response_dict)
        return response_dict

    def convert(self, response_dict: Dict[str, Any]):
        """
        Convert chat response message to a format expected by external clients:
        :param response_dict: chat response message to be sent out
        """
        # Ensure that we return ChatMessageType as a string in output json
        message_dict: Dict[str, Any] = response_dict.get('response', None)
        if message_dict is not None:
            self.convert_message(message_dict)

    def convert_message(self, message_dict: Dict[str, Any]):
        """
        Convert chat message to a format expected by external clients:
        :param message_dict: chat message to process
        """
        # Ensure that we return ChatMessageType as a string in output json
        response_type = message_dict.get('type', None)
        if response_type is not None:
            message_dict['type'] =\
                ChatMessageType.from_response_type(response_type).name
        chat_context: Dict[str, Any] = message_dict.get('chat_context', None)
        if chat_context is not None:
            for chat_history in chat_context.get("chat_histories", []):
                for chat_message in chat_history.get("messages", []):
                    self.convert_message(chat_message)

    def from_dict(self, obj_dict: Dict[str, object]) -> object:
        """
        :param obj_dict: The data-only dictionary to be converted into an object
        :return: An object instance created from the given dictionary.
                If obj_dict is None, the returned object should also be None.
                If obj_dict is not the correct type, it is also reasonable
                to return None.
        """
        raise NotImplementedError

    JSON_PRIMITIVES = (str, int, float, bool, type(None))
    SKIP_TYPES = (ChatMessageType,)

    def to_json_safe(self, obj: Any):
        """
        Recursively convert obj into a JSON-serializable structure.
        Don't touch the values of SKIP_TYPES.
        Non-serializable values are replaced with None.
        """
        # Fast path: JSON primitives or types we need to skip:
        if isinstance(obj, (self.SKIP_TYPES, self.JSON_PRIMITIVES)):
            return obj

        # Enums → their value (or None if value is not primitive)
        if isinstance(obj, Enum):
            if isinstance(obj.value, self.JSON_PRIMITIVES):
                return obj.value
            return None

        # Dict → keys must be strings, values recursively processed
        if isinstance(obj, dict):
            safe_dict = {}
            for k, v in obj.items():
                if isinstance(k, str):
                    safe_dict[k] = self.to_json_safe(v)
                # Non-string keys are dropped
            return safe_dict

        # List / tuple → list
        if isinstance(obj, (list, tuple)):
            return [self.to_json_safe(v) for v in obj]

        # Everything else → None
        return None




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.filters.message_filter import MessageFilter
from neuro_san.internals.messages.chat_message_type import ChatMessageType


class AnswerMessageFilter(MessageFilter):
    """
    MessageFilter implementation for a message with "the answer" in it.
    """

    def allow_message(self, chat_message_dict: Dict[str, Any], message_type: ChatMessageType) -> bool:
        """
        Determine whether to allow the message through.

        :param chat_message_dict: The ChatMessage dictionary to process.
        :param message_type: The ChatMessageType of the chat_message_dictionary to process.
        :return: True if the message should be allowed through to the client. False otherwise.
        """
        if message_type not in (ChatMessageType.AI, ChatMessageType.AGENT_FRAMEWORK):
            # Final answers are only ever AI or AgentFramework Messages
            return False

        origin: List[Dict[str, Any]] = chat_message_dict.get("origin")
        if origin is not None and len(origin) > 1:
            # Final answers only come from the FrontMan,
            # whose origin length is the only one of length 1.
            return False

        text: str = chat_message_dict.get("text")
        structure: Dict[str, Any] = chat_message_dict.get("structure")
        if text is None and structure is None:
            # Final answers need to be text or structure.
            # There might be more options in the future.
            return False

        # Meets all our criteria. Let it through.
        return True




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

from neuro_san.interfaces.usage_logger import UsageLogger


class DebugUsageLogger(UsageLogger):
    """
    Implementation of the UsageLogger interface that merely spits out
    usage stats to the logger.
    """

    async def log_usage(self, token_dict: Dict[str, Any], request_metadata: Dict[str, Any]):
        """
        Logs the token usage for external capture.

        :param token_dict: A dictionary that describes overall token usage for a completed request.

                For each class of LLM (more or less equivalent to an LLM provider), there will
                be one key whose value is a dictionary with some other keys:

                Relevant keys include:
                    "completion_tokens" - Integer number of tokens generated in response to LLM input
                    "prompt_tokens" - Integer number of tokens that provide input to an LLM
                    "time_taken_in_seconds" - Float describing the total wall-clock time taken for the request.
                    "total_cost" -  An estimation of the cost in USD of the request.
                                    This number is to be taken with a grain of salt, as these estimations
                                    can come from model costs from libraries instead of directly from
                                    providers.
                    "total_tokens" - Total tokens used for the request.

                More keys can appear, but should not be counted on.
                The ones listed above contain potentially salient information for usage logging purposes.

        :param request_metadata: A dictionary of filtered request metadata whose keys contain
                identifying information for the usage log.
        """




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Union

import json
import uuid

from copy import copy
from logging import Logger
from logging import getLogger

from pydantic_core import ValidationError

from langchain.agents.factory import create_agent
from langchain.agents.middleware.types import AgentMiddleware
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.messages.base import BaseMessage
from langchain_core.messages.human import HumanMessage
from langchain_core.messages.system import SystemMessage
from langchain_core.runnables.base import Runnable
from langchain_core.runnables.passthrough import RunnablePassthrough
from langchain_core.tools.base import BaseTool

from neuro_san.internals.errors.error_detector import ErrorDetector
from neuro_san.internals.interfaces.context_type_llm_factory import ContextTypeLlmFactory
from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.journals.journal import Journal
from neuro_san.internals.journals.intercepting_journal import InterceptingJournal
from neuro_san.internals.journals.originating_journal import OriginatingJournal
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.messages.agent_tool_result_message import AgentToolResultMessage
from neuro_san.internals.messages.base_message_dictionary_converter import BaseMessageDictionaryConverter
from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.run_context import RunContext
from neuro_san.internals.run_context.interfaces.tool_caller import ToolCaller
from neuro_san.internals.run_context.langchain.core.base_tool_factory import BaseToolFactory
from neuro_san.internals.run_context.langchain.core.langchain_run import LangChainRun
from neuro_san.internals.run_context.langchain.core.run_context_runnable import RunContextRunnable
from neuro_san.internals.run_context.langchain.llms.langchain_llm_resources import LangChainLlmResources
from neuro_san.internals.run_context.langchain.middleware.middleware_factory import MiddlewareFactory


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class LangChainRunContext(RunContext):
    """
    LangChain implementation on RunContext interface supporting high-level LLM usage
    This ends up being useful:
        https://python.langchain.com/docs/modules/tools/tools_as_openai_functions/
    Note that "tools" can be just a list of OpenAI functions JSON.
    """

    # pylint: disable=too-many-arguments, too-many-positional-arguments
    def __init__(self, llm_config: Dict[str, Any],
                 parent_run_context: RunContext,
                 tool_caller: ToolCaller,
                 invocation_context: InvocationContext,
                 chat_context: Dict[str, Any],
                 middleware_config: List[Dict[str, Any]] = None):
        """
        Constructor

        :param llm_config: The default llm_config to use as an overlay
                            for the tool-specific llm_config
        :param parent_run_context: The parent RunContext that is calling this one. Can be None.
        :param tool_caller: The tool caller to use
        :param invocation_context: The context policy container that pertains to the invocation
                    of the agent.
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        :param middleware_config: An ordered list of middleware configurations
        """
        self.chat_history: List[BaseMessage] = []
        self.middleware_config: List[Dict[str, Any]] = middleware_config
        self.journal: Journal = None
        self.interceptor: InterceptingJournal = None
        self.llm_resources: LangChainLlmResources = None
        self.agent_chain: Runnable = None

        # This might get modified in create_resources() (for now)
        self.llm_config: Dict[str, Any] = llm_config
        self.run_id_base: str = str(uuid.uuid4())

        self.tools: List[BaseTool] = []
        self.error_detector: ErrorDetector = None
        self.recent_human_message: HumanMessage = None
        self.tool_caller: ToolCaller = tool_caller
        self.invocation_context: InvocationContext = invocation_context
        self.chat_context: Dict[str, Any] = chat_context
        self.origin: List[Dict[str, Any]] = []
        # Have we already created resources for this RunContext:
        self.resources_created: bool = False
        # Default logger
        self.logger: Logger = getLogger(self.__class__.__name__)

        parent_origin: List[Dict[str, Any]] = []
        if parent_run_context is not None:

            # Get other stuff from parent if not specified
            if self.invocation_context is None:
                self.invocation_context = parent_run_context.get_invocation_context()
            if self.chat_context is None:
                self.chat_context = parent_run_context.get_chat_context()
            parent_origin = parent_run_context.get_origin()

            # Initialize the origin.
            agent_name: str = tool_caller.get_name()
            origination: Origination = self.invocation_context.get_origination()
            self.origin = origination.add_spec_name_to_origin(parent_origin, agent_name)

        self.update_from_chat_context(self.chat_context)

        # Set up so local logging gives origin info.
        if self.origin is not None and len(self.origin) > 0:
            full_name: str = Origination.get_full_name_from_origin(self.origin)
            self.logger = getLogger(full_name)

        if self.invocation_context is not None:
            # Sets up self.journal
            self.update_invocation_context(self.invocation_context)

    async def create_resources(self, agent_name: str,
                               instructions: str,
                               assignments: str,
                               tool_names: List[str] = None):
        """
        Creates resources for later use within the RunContext instance.
        Results are stored as a member in this instance for future use.

        Note that even though this method is labeled as async, we don't
        really do any async method calls in here for this implementation.

        :param agent_name: String name of the agent
        :param instructions: string instructions that are used
                    to create the agent
        :param assignments: string assignments of function parameters that are used as input
        :param tool_names: The list of registered tool names to use.
                    Default is None implying no tool is to be called.
        """
        # DEF - Remove the arg if possible
        if self.resources_created:
            # Don't create RunContext resources twice -
            # we could possibly leak some run-time resources.
            return

        _ = agent_name

        full_name: str = Origination.get_full_name_from_origin(self.origin)
        agent_spec: Dict[str, Any] = self.tool_caller.get_agent_tool_spec()

        # Now that we have a name, we can create an ErrorDetector for the output.
        self.error_detector = ErrorDetector(full_name,
                                            error_formatter_name=agent_spec.get("error_formatter"),
                                            system_error_fragments=["Agent stopped"],
                                            agent_error_fragments=agent_spec.get("error_fragments"))

        if tool_names is not None:
            factory = BaseToolFactory(self.tool_caller, self.invocation_context, self.journal)
            for tool_name in tool_names:
                tool: Union[BaseTool | List[BaseTool]] = await factory.create_base_tool(tool_name)
                if tool is not None:
                    if isinstance(tool, List):
                        self.tools.extend(tool)
                    else:
                        self.tools.append(tool)

        if not self.chat_history:
            # Write instructions as the first message in the journal.
            # "instructions" is provided to the agent in create_agent() and is not in the chat history.
            await self.journal.write_message(SystemMessage(instructions))

        self.agent_chain = self.create_agent_with_fallbacks(instructions)
        self.resources_created = True

    def create_agent_with_fallbacks(self, instructions: str) -> Runnable:
        """
        Creates an agent with potential fallback llms to use.
        :param instructions: The instructions to use for the agent
        :return: An Agent (Runnable)
        """
        # Initialize our return value
        agent: Runnable = None

        # Get the factory we will use
        llm_factory: ContextTypeLlmFactory = self.invocation_context.get_llm_factory()

        # Prepare a list of fallbacks.  By default, the llm_config itself is a single-entry fallback list.
        fallbacks: List[Dict[str, Any]] = [self.llm_config]
        fallbacks = self.llm_config.get("fallbacks", fallbacks)

        # Initialize a list of chain fallbacks. This may or may not get filled.
        chain_fallbacks: List[Runnable] = []

        # Go through the list of fallbacks in the config.
        for index, fallback in enumerate(fallbacks):

            # Create a model we might use.
            one_llm_resources: LangChainLlmResources = llm_factory.create_llm(fallback)
            one_agent: Runnable = self.create_agent(instructions, one_llm_resources.get_model())

            if index == 0:
                # The first agent is the one we want to be our main guy.
                agent = one_agent
                # For now. Could be problems with different providers w/ token counting.
                self.llm_resources = one_llm_resources
            else:
                # Anything later than the first guy is considered a fallback. Add it to the list.
                chain_fallbacks.append(one_agent)

        if len(chain_fallbacks) > 0:
            # Set up fallbacks.
            # See https://python.langchain.com/docs/how_to/tools_error/#tryexcept-tool-call
            agent = agent.with_fallbacks(chain_fallbacks)

        return agent

    def create_agent(self, instructions: str, llm: BaseLanguageModel) -> Runnable:
        """
        Creates an agent.
        :param instructions: The instructions to use for the agent
        :param llm: The BaseLanguageModel to use for the agent
        :return: An Agent (Runnable)
        """

        # Create any middleware instances that were specified, in the order they were specified.
        # This will be None for most simple situations.
        middleware_factory = MiddlewareFactory(self.invocation_context, self.origin, self.chat_history)
        sly_data: Dict[str, Any] = self.tool_caller.get_sly_data()

        middleware: List[AgentMiddleware] = None
        checkpointer: Any = None
        middleware, checkpointer = middleware_factory.create_agent_middleware(self.middleware_config, sly_data)

        return create_agent(
            model=llm,
            tools=self.tools,
            middleware=middleware,
            checkpointer=checkpointer,
            system_prompt=instructions,
        )

    async def submit_message(self, user_message: str) -> Run:
        """
        Submits a message to the model used by this instance.

        Note that even though this method is labeled as async, we don't
        really do any async method calls in here for this implementation.

        :param user_message: The message to submit
        :return: The run which is processing the agent's message
        """
        # Contruct a human message out of the text of the user message
        # Don't add this to the chat history yet.
        try:
            self.recent_human_message = HumanMessage(user_message)
        except ValidationError as exception:
            full_name: str = Origination.get_full_name_from_origin(self.origin)
            message = f"ValidationError in {full_name} with message: {user_message}"
            raise ValueError(message) from exception

        # Create a run to return
        run = LangChainRun(self.run_id_base, self.chat_history)
        return run

    async def wait_on_run(self, run: Run, journal: Journal = None) -> Run:
        """
        Loops on the given run's status for model invokation.

        This truly is an asynchronous method.

        :param run: The run to wait on
        :param journal: The Journal which captures the "thinking" messages.
        :return: An potentially updated run
        """
        _ = run, journal

        # Chat history is updated in write_message() below, so to save on
        # some tokens, make a shallow copy of it here as we send it to the LLM
        previous_chat_history: List[BaseMessage] = copy(self.chat_history)

        inputs = {
            # All agents have a "messages" field, which is a list of messages.
            # To invoke the agent, pass a new list containing the previous chat history
            # along with the new user message.
            # https://docs.langchain.com/oss/python/langchain/agents#invocation
            "messages": previous_chat_history + [self.recent_human_message]
        }

        # Chat history is updated in write_message
        await self.journal.write_message(self.recent_human_message)

        run: Run = LangChainRun(self.run_id_base, self.chat_history)
        session_id: str = run.get_id()

        runnable = RunContextRunnable(agent_chain=self.agent_chain,
                                      primary_llm=self.llm_resources.get_model(),
                                      invocation_context=self.invocation_context,
                                      journal=self.journal,
                                      interceptor=self.interceptor,
                                      origin=self.origin,
                                      tool_caller=self.tool_caller,
                                      error_detector=self.error_detector,
                                      session_id=session_id)
        runnable_config: Dict[str, Any] = runnable.prepare_runnable_config(session_id=session_id, use_run_name=True)

        # This needs to be run as a chain otherwise LangSmith will pick up two
        # trace names for the same request.
        chain: Runnable = RunnablePassthrough() | runnable

        await chain.ainvoke(input=inputs, config=runnable_config)

        return run

    async def get_response(self) -> List[BaseMessage]:
        """
        :return: The list of messages from the instance's thread.
        """
        # Not sure if this is the right thing, as this will be langchain-y stuff.
        return self.chat_history

    async def submit_tool_outputs(self, run: Run, tool_outputs: List[Dict[str, Any]]) -> Run:
        """
        :param run: The Run handling the execution of the agent
        :param tool_outputs: The tool outputs to submit
                The component dictionaries can have the following keys:
                    "origin"        A List of origin dictionaries indicating the origin of the run.
                    "output"        A string representing the output of the tool call
                    "sly_data"      Optional sly_data dictionary that might have returned from an external tool.
                    "tool_call_id"  The string id of the tool_call being executed
        :return: A potentially updated run handle
        """
        tool_message: BaseMessage = None
        if tool_outputs is not None and len(tool_outputs) > 0:
            for tool_output in tool_outputs:
                tool_message = self.parse_tool_output(tool_output)
                if tool_message is not None:
                    # Write the tool message to the journal.
                    # Chat history should only contain user and AI messages.
                    # It will not be updated in the write_message() call below.
                    await self.journal.write_message(tool_message)

        # Create a run to return
        run = LangChainRun(self.run_id_base, self.chat_history, tool_message=tool_message)

        return run

    def parse_tool_output(self, tool_output: Dict[str, Any]) -> BaseMessage:
        """
        Parse a single tool_output dictionary for its results
        :return: A message representing the output from the tool.
        """

        # Get a Message for each output and add to the chat history.
        # Assuming dictionary
        tool_chat_list_string = tool_output.get("output", None)
        if tool_chat_list_string is None:
            # Dunno what to do with None tool output
            return None
        if isinstance(tool_chat_list_string, tuple):
            # Sometimes output comes back as a tuple.
            # The output we want is the first element of the tuple.
            tool_chat_list_string = tool_chat_list_string[0]

        tool_message: BaseMessage = None
        if isinstance(tool_chat_list_string, str):
            # Sometimes output comes back as a list.
            # The output we want is the first element of the list.
            tool_message = self.parse_tool_chat_list_string(tool_chat_list_string, tool_output.get("origin"))

        elif isinstance(tool_chat_list_string, BaseMessage):
            tool_message = AgentToolResultMessage(content=tool_chat_list_string.content,
                                                  tool_result_origin=tool_output.get("origin"))

        elif isinstance(tool_chat_list_string, list) and isinstance(tool_chat_list_string[-1], BaseMessage):
            # Always take the last element of the list as the answer to the tool output
            last_message: BaseMessage = tool_chat_list_string[-1]
            tool_message = AgentToolResultMessage(content=last_message.content,
                                                  tool_result_origin=tool_output.get("origin"))
        else:
            self.logger.warning("Dunno what to do with %s tool output %s",
                                str(tool_chat_list_string.__class__.__name__),
                                str(tool_chat_list_string))
            return None

        # Integrate any sly data
        tool_sly_data: Dict[str, Any] = tool_output.get("sly_data")
        if tool_sly_data and tool_sly_data != self.tool_caller.get_sly_data():
            # We have sly data from the tool output that is not the same as our own
            # and it has data in it.  Integrate that.
            # It's possible we might need to run a SlyDataRedactor against from_download.sly_data on this.
            self.tool_caller.get_sly_data().update(tool_sly_data)

        return tool_message

    def parse_tool_chat_list_string(self, tool_chat_list_string: str, origin: str) -> BaseMessage:
        """
        Parse a tool output string into a list of messages
        :param tool_chat_list_string: The string to parse
        :param origin: The origin of the tool
        :return: A list of messages representing the output from the tool.
        """

        # Remove bracketing quotes from within the string
        while (tool_chat_list_string[0] == '"' and tool_chat_list_string[-1] == '"') or \
              (tool_chat_list_string[0] == "'" and tool_chat_list_string[-1] == "'"):
            tool_chat_list_string = tool_chat_list_string[1:-1]

        # Remove escaping
        tool_chat_list_string = tool_chat_list_string.replace('\\"', '"')
        # Put back some escaping of double quotes in messages that are not json.
        # We have to do this because gpt-4o seems to not like json braces in its
        # input, but now we have to deal with the consequences in the output.
        # See ArgumentAssigner.get_args_value_as_string().
        tool_chat_list_string = tool_chat_list_string.replace('\\"', '\\\\\"')

        # Decode the JSON in that string now.
        tool_chat_list: List[Dict[str, Any]] = None
        try:
            tool_chat_list = json.loads(tool_chat_list_string)
        except json.decoder.JSONDecodeError as exception:
            self.logger.error("Exception: %s parsing %s", str(exception), str(tool_chat_list_string))
            raise exception

        # The tool_output seems to contain the entire chat history of
        # the call to the tool. For now just take the last one as the answer.
        tool_result_dict = tool_chat_list[-1]

        # Turn that guy into a BaseMessage
        # You might expect that this should be a ToolMessage, but making that
        # kind of conversion at this point runs into problems with OpenAI models
        # that process them.  So, to make things continue to work, report the
        # content as an AI message - as if the bot came up with the answer itself.
        tool_message = AgentToolResultMessage(content=tool_result_dict.get("content"),
                                              tool_result_origin=origin)

        return tool_message

    async def delete_resources(self, parent_run_context: RunContext = None):
        """
        Cleans up the service-side resources associated with this instance
        :param parent_run_context: A parent RunContext perhaps the same instance,
                        but perhaps not.  Default is None
        """

        # Release model related resources:
        if self.llm_resources:
            await self.llm_resources.delete_resources()

        self.tools = []
        self.chat_history = []
        self.agent_chain = None
        self.recent_human_message = None
        self.llm_resources = None
        self.journal = None
        self.interceptor = None

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        if self.tool_caller is None:
            return None

        return self.tool_caller.get_agent_tool_spec()

    def get_invocation_context(self) -> InvocationContext:
        """
        :return: The InvocationContext policy container that pertains to the invocation
                    of the agent.
        """
        return self.invocation_context

    def get_chat_context(self) -> Dict[str, Any]:
        """
        :return: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
                Can be None when a new conversation has been started.
        """
        return self.chat_context

    def get_origin(self) -> List[Dict[str, Any]]:
        """
        :return: A List of origin dictionaries indicating the origin of the run.
                The origin can be considered a path to the original call to the front-man.
                Origin dictionaries themselves each have the following keys:
                    "tool"                  The string name of the tool in the spec
                    "instantiation_index"   An integer indicating which incarnation
                                            of the tool is being dealt with.
        """
        return self.origin

    def update_invocation_context(self, invocation_context: InvocationContext):
        """
        Update internal state based on the InvocationContext instance passed in.
        :param invocation_context: The context policy container that pertains to the invocation
        """
        old_interceptor: InterceptingJournal = self.interceptor
        self.invocation_context = invocation_context

        # Make a nested chain where each journal is wrapped by the next
        base_journal: Journal = self.invocation_context.get_journal()
        self.interceptor = InterceptingJournal(wrapped_journal=base_journal, origin=self.origin)

        # The SystemMessage has already been written to the journal
        # need to transfer it over when this shift happens.
        if old_interceptor is not None:
            for message in old_interceptor.get_messages():
                self.interceptor.write_unwrapped_message(message, self.origin)
        self.journal = OriginatingJournal(self.interceptor, self.origin, self.chat_history)

    def update_from_chat_context(self, chat_context: Dict[str, Any]):
        """
        :param chat_context: A ChatContext dictionary that contains all the state necessary
                to carry on a previous conversation, possibly from a different server.
        """
        self.chat_context = chat_context

        if self.chat_context is None:
            return

        # See if our origin appears in the chat histories.
        # If so, get ours from there.
        empty: List[Any] = []
        chat_histories: List[Dict[str, Any]] = self.chat_context.get("chat_histories", empty)
        our_origin_str: str = Origination.get_full_name_from_origin(self.origin)
        for one_chat_history in chat_histories:

            # See if the origin matches our own
            test_origin: List[Dict[str, Any]] = one_chat_history.get("origin", empty)
            test_origin_str: str = Origination.get_full_name_from_origin(test_origin)
            if test_origin_str != our_origin_str:
                continue

            one_messages: List[Dict[str, Any]] = one_chat_history.get("messages", empty)
            if not one_messages:
                # Empty list - Nothing to convert. Use default empty list.
                break

            converter = BaseMessageDictionaryConverter()
            self.chat_history = []
            for chat_message in one_messages:
                base_message: BaseMessage = converter.from_dict(chat_message)
                if base_message is not None:
                    self.chat_history.append(base_message)

            # Nothing left to search for
            break

    def get_journal(self) -> Journal:
        """
        :return: The Journal associated with the instance
        """
        return self.journal




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import List

from neuro_san.interfaces.concierge_session import ConciergeSession
from neuro_san.internals.network_providers.agent_network_storage import AgentNetworkStorage
from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler
from neuro_san.session.direct_concierge_session import DirectConciergeSession


class ConciergeHandler(BaseRequestHandler):
    """
    Handler class for neuro-san "concierge" API call.
    """

    async def get(self):
        """
        Implementation of GET request handler for "concierge" API call.
        """
        metadata: Dict[str, Any] = self.get_metadata()
        self.application.start_client_request(metadata, "/api/v1/list")
        network_storage_dict: Dict[str, AgentNetworkStorage] = self.server_context.get_network_storage_dict()
        public_storage: AgentNetworkStorage = network_storage_dict.get("public")

        # See what the authorizer says
        allowed_agents: List[str] = await self.agent_policy.list_agents(metadata)

        try:
            data: Dict[str, Any] = {}
            session: ConciergeSession = DirectConciergeSession(public_storage, metadata=metadata)
            result_dict: Dict[str, Any] = session.list(data)

            # Maybe remove agents if the agent_policy has something to say.
            if allowed_agents is not None:
                result_dict = self.pare_allowed_agents(allowed_agents, result_dict)

            # Return response to the HTTP client
            self.set_header("Content-Type", "application/json")
            self.write(result_dict)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            self.do_finish()
            self.application.finish_client_request(metadata, "/api/v1/list")

    def pare_allowed_agents(self, allowed_agents: List[str], result_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        Remove agents which are not allowed.

        :param allowed_agents: A list of agent names that are allowed by the authorization system
        :param result_dict: A dictionary version of the ConciergeResponse
                protobuf structure. Has the following keys:
            "agents" - the sequence of dictionaries describing available agents
        :return: A dictionary version of the ConciergeResponse
                protobuf structure. Has the following keys:
            "agents" - the sequence of dictionaries describing available agents
        """

        empty: List[Dict[str, Any]] = []
        agent_infos: List[Dict[str, Any]] = result_dict.get("agents", empty)

        # Create a dictionary of agent names for quick lookup
        agent_info_dict: Dict[str, Dict[str, Any]] = {}
        for agent_info in agent_infos:
            agent_name: str = agent_info.get("agent_name")
            if agent_name in allowed_agents:
                agent_info_dict[agent_name] = agent_info

        # Recreate the list of agents in the results by taking the values from the dictionary
        result_dict["agents"] = list(agent_info_dict.values())
        return result_dict




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from logging import Logger
from logging import getLogger
from os import environ

from leaf_common.config.resolver_util import ResolverUtil

from neuro_san.internals.authorization.interfaces.authorizer import Authorizer
from neuro_san.internals.authorization.null.always_yes_authorizer import AlwaysYesAuthorizer


class AuthorizerFactory:
    """
    Factory class for getting the appropriate Authorizer instance
    """

    @staticmethod
    def create_authorizer() -> Authorizer:
        """
        Factory method for creating the appropriate Authorizer instance
        """
        authorizer: Authorizer = None

        auth_classname: str = environ.get("AGENT_AUTHORIZER")
        if auth_classname is not None and len(auth_classname) > 0:
            # Note that this will raise an exception if the class is not found
            authorizer = ResolverUtil.create_instance(auth_classname, "AGENT_AUTHORIZER env var", Authorizer)
        else:
            authorizer = AlwaysYesAuthorizer()

        # We don't leave the room unless we have what we came for
        if authorizer is None:
            raise ValueError(f"Unable to create an Authorizer instance from env var AGENT_AUTHORIZER: {auth_classname}")

        logger: Logger = getLogger(__name__)
        logger.info("Using Authorizer: %s.%s", authorizer.__class__.__module__, authorizer.__class__.__name__)

        return authorizer




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any
from typing import Dict

import json
import requests

from neuro_san.interfaces.concierge_session import ConciergeSession
from neuro_san.session.abstract_http_service_agent_session import AbstractHttpServiceAgentSession


class HttpConciergeSession(AbstractHttpServiceAgentSession, ConciergeSession):
    """
    Implementation of ConciergeSession that talks to an HTTP service.
    This is largely only used by command-line tests.
    """

    def list(self, request_dict: Dict[str, Any]) -> Dict[str, Any]:
        """
        :param request_dict: A dictionary version of the ConciergeRequest
                    protobuf structure. Has the following keys:
                        <None>
        :return: A dictionary version of the ConciergeResponse
                    protobuf structure. Has the following keys:
                "agents" - the sequence of dictionaries describing available agents
        """
        path: str = self.get_request_path("list")
        try:
            response = requests.get(path, json=request_dict, headers=self.get_headers(),
                                    timeout=self.timeout_in_seconds)
            result_dict = json.loads(response.text)
            return result_dict
        except Exception as exc:  # pylint: disable=broad-exception-caught
            raise ValueError(self.help_message(path)) from exc




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List

from neuro_san.internals.messages.origination import Origination


class ToolArgumentReporting:
    """
    Utility class to assist in preparing arguments dictionaries when reporing starting a tool.
    """

    # List of keys for policy objects that cannot be serialized in a message.
    # The first two are set in AbstractClassActivation.
    # The last one is injected by Langgraph when using MCP tools with langchain-mcp-adapters>=0.1.14.
    POLICY_OBJECT_KEYS: List[str] = ["reservationist", "progress_reporter", "runtime"]

    @staticmethod
    def prepare_tool_start_dict(tool_args: Dict[str, Any],
                                origin: List[Dict[str, Any]] = None) -> Dict[str, Any]:
        """
        Common code to prepare a tool start dictionary.

        :param tool_args: The arguments that will be passed to the tool
        :param origin: A List of origin dictionaries indicating the origin of the run.
        :return: A dictionary for a future journal entry
        """

        modified_tool_args: Dict[str, Any] = tool_args.copy()

        # Combine the original tool tool_args with origin metadata, if available.
        if origin is not None:
            modified_tool_args["origin"] = origin

            full_name: str = Origination.get_full_name_from_origin(origin)
            modified_tool_args["origin_str"] = full_name

        # Remove policy object keys from the args that cannot be serialized in a message.
        for key in ToolArgumentReporting.POLICY_OBJECT_KEYS:
            if key in modified_tool_args:
                del modified_tool_args[key]

        # Create a dictionary for a future journal entry for this invocation
        tool_start_dict: Dict[str, Any] = {
            "tool_start": True,
            "tool_args": modified_tool_args
        }

        return tool_start_dict




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict

from logging import getLogger
from logging import Logger

from leaf_common.config.config_filter import ConfigFilter


class ServedManifestConfigFilter(ConfigFilter):
    """
    Implementation of the ConfigFilter interface that reads the contents
    of a single manifest file for agent networks/registries, removing any entries
    that are not supposed to be served.
    """

    def __init__(self, manifest_file: str, warn_on_skip: bool = True, entry_for_skipped: bool = False):
        """
        Constructor

        :param manifest_file: The name of the manifest file we are processing for logging purposes
        :param warn_on_skip: If True, a warning will be logged if an entry is skipped
        :param entry_for_skipped: If True, the skipped entry will be included in the keys of the
                                output dictionary but the value will be None.
        """
        self.logger: Logger = getLogger(self.__class__.__name__)
        self.manifest_file: str = manifest_file
        self.warn_on_skip: bool = warn_on_skip
        self.entry_for_skipped: bool = entry_for_skipped

    def filter_config(self, basis_config: Dict[str, Dict[str, Any]]) -> Dict[str, Dict[str, Any]]:
        """
        Filters the given basis config.

        :param basis_config: The config dictionary to act as the basis
                for filtering
        :return: A config dictionary, potentially modified as per the
                policy encapsulated by the implementation
        """

        filtered: Dict[str, Dict[str, Any]] = {}

        for key, value in basis_config.items():

            skip: bool = False
            if value is None:
                skip = True
            elif isinstance(value, dict) and not value.get("serve", False):
                skip = True

            if skip:
                if self.warn_on_skip:
                    self.logger.warning("Manifest entry for %s in file %s will not be served, " +
                                        "per the 'serve' key in its config (default is False). Skipping.",
                                        key, self.manifest_file)
                # Instead of merely omitting the entry, we set it to None.
                # This allows for multiple manifest overlays to work.
                if self.entry_for_skipped:
                    filtered[key] = value
            else:
                filtered[key] = value

        return filtered




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any


class RunTarget:
    """
    Interface for something that can be run as part of the course of a streaming_chat.
    """

    async def run_it(self, inputs: Any) -> Any:
        """
        Entry point method for the run.

        :param inputs: A list of inputs from the user.
        :return: The outputs of the run.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

import os
import tempfile
import unittest

from unittest.mock import patch

from neuro_san.client.hocon_validator_cli import HoconValidatorCli


class TestHoconValidatorCli(unittest.TestCase):
    """
    Unit tests for the HoconValidatorCli class.
    """

    def setUp(self):
        """Set up test fixtures."""
        self.test_dir = os.path.dirname(os.path.abspath(__file__))
        self.neuro_san_dir = os.path.dirname(os.path.dirname(self.test_dir))
        self.registries_dir = os.path.join(self.neuro_san_dir, "registries")

    def test_valid_hocon_returns_zero(self):
        """Test that a valid HOCON file returns exit code 0."""
        hocon_file = os.path.join(self.registries_dir, "hello_world.hocon")
        with patch("sys.argv", ["hocon_validator_cli", hocon_file]):
            cli = HoconValidatorCli()
            exit_code = cli.main()
        self.assertEqual(exit_code, 0)

    def test_file_not_found_returns_two(self):
        """Test that a non-existent file returns exit code 2."""
        with patch("sys.argv", ["hocon_validator_cli", "/nonexistent/file.hocon"]):
            cli = HoconValidatorCli()
            exit_code = cli.main()
        self.assertEqual(exit_code, 2)

    def test_missing_agent_returns_one(self):
        """Test that a HOCON with missing agent reference returns exit code 1."""
        hocon_content = '''
        {
            tools: [
                {
                    name: "test_agent"
                    instructions: "Test instructions"
                    tools: ["nonexistent_tool"]
                }
            ]
        }
        '''
        with tempfile.NamedTemporaryFile(mode="w", suffix=".hocon", delete=False) as f:
            f.write(hocon_content)
            temp_file = f.name
        try:
            with patch("sys.argv", ["hocon_validator_cli", temp_file, "--registry-dir", self.neuro_san_dir]):
                cli = HoconValidatorCli()
                exit_code = cli.main()
            self.assertEqual(exit_code, 1)
        finally:
            os.unlink(temp_file)

    def test_unreachable_agent_returns_one(self):
        """Test that a HOCON with unreachable agent returns exit code 1."""
        hocon_content = '''
        {
            tools: [
                {
                    name: "main_agent"
                    instructions: "Main agent"
                    tools: []
                },
                {
                    name: "orphan_agent"
                    instructions: "This agent is never referenced"
                    tools: []
                }
            ]
        }
        '''
        with tempfile.NamedTemporaryFile(mode="w", suffix=".hocon", delete=False) as f:
            f.write(hocon_content)
            temp_file = f.name
        try:
            with patch("sys.argv", ["hocon_validator_cli", temp_file, "--registry-dir", self.neuro_san_dir]):
                cli = HoconValidatorCli()
                exit_code = cli.main()
            self.assertEqual(exit_code, 1)
        finally:
            os.unlink(temp_file)

    def test_external_agents_flag(self):
        """Test that --external-agents flag allows external agent references."""
        hocon_content = '''
        {
            tools: [
                {
                    name: "test_agent"
                    instructions: "Test instructions"
                    tools: ["/external_agent"]
                }
            ]
        }
        '''
        with tempfile.NamedTemporaryFile(mode="w", suffix=".hocon", delete=False) as f:
            f.write(hocon_content)
            temp_file = f.name
        try:
            with patch("sys.argv", ["hocon_validator_cli", temp_file,
                                    "--registry-dir", self.neuro_san_dir,
                                    "--external-agents", "/external_agent"]):
                cli = HoconValidatorCli()
                exit_code = cli.main()
            self.assertEqual(exit_code, 0)
        finally:
            os.unlink(temp_file)

    def test_verbose_flag(self):
        """Test that --verbose flag works without errors."""
        hocon_file = os.path.join(self.registries_dir, "hello_world.hocon")
        with patch("sys.argv", ["hocon_validator_cli", hocon_file, "--verbose"]):
            cli = HoconValidatorCli()
            exit_code = cli.main()
        self.assertEqual(exit_code, 0)


if __name__ == "__main__":
    unittest.main()




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT

from typing import Any

from neuro_san.test.evaluators.abstract_agent_evaluator import AbstractAgentEvaluator


class ValueAgentEvaluator(AbstractAgentEvaluator):
    """
    AbstractAgentEvaluator implementation that looks for specific values in output.
    """

    def test_one(self, verify_value: Any, test_value: Any):
        """
        :param verify_value: The value to verify against
        :param test_value: The value appearing in the test sample
        """
        if self.negate:
            self.asserts.assertNotEqual(verify_value, test_value)
        else:
            self.asserts.assertEqual(verify_value, test_value)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict


class AgentSpecProvider:
    """
    Interface for an entity to return a copy of the agent spec that pertains to it.
    """

    def get_agent_tool_spec(self) -> Dict[str, Any]:
        """
        :return: the dictionary describing the data-driven agent
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import os

from logging import Logger
from logging import getLogger

from pydantic import ConfigDict
from typing_extensions import override

from langchain_core.callbacks.base import BaseCallbackHandler
from langchain_core.messages.base import BaseMessage
from langchain_core.messages.base import messages_to_dict
from langchain_core.runnables.base import Other
from langchain_core.runnables.base import RunnableConfig
from langchain_core.runnables.passthrough import RunnablePassthrough
from langchain_core.runnables.utils import Input
from langchain_core.runnables.utils import Output

from neuro_san.internals.interfaces.invocation_context import InvocationContext
from neuro_san.internals.interfaces.run_target import RunTarget
from neuro_san.internals.journals.intercepting_journal import InterceptingJournal
from neuro_san.internals.messages.origination import Origination
from neuro_san.internals.utils.metadata_util import MetadataUtil


class NeuroSanRunnable(RunnablePassthrough, RunTarget):
    """
    RunnablePassthrough implementation that intercepts journal messages
    for a particular origin.
    """

    # Declarations of member variables here satisfy Pydantic style,
    # which is a type validator that langchain is based on which
    # is able to use JSON schema definitions to validate fields.
    invocation_context: InvocationContext = None

    interceptor: InterceptingJournal = None

    origin: List[Dict[str, Any]] = None

    session_id: str = None

    # Default logger
    logger: Optional[Logger] = None

    run_target: Optional[RunTarget] = None

    # This guy needs to be a pydantic class and in order to have
    # any non-pydantic non-serializable members, we need to do this.
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Constructor
        """
        # Use the run_target from the kwargs as the afunc parameter
        # for the RunnablePassthrough.  If that doesn't exist, assume
        # our own run_it() method is overridden and use that.
        run_target: RunTarget = kwargs.get("run_target", self)
        super().__init__(afunc=run_target.run_it, **kwargs)

        self.logger: Logger = getLogger(self.__class__.__name__)

    # pylint: disable=redefined-builtin
    @override
    async def ainvoke(
        self,
        input: Other,
        config: RunnableConfig | None = None,
        **kwargs: Any | None,
    ) -> Other:

        # Called by langchain infrastruture when a chain is invoked.
        # Calling the super here means that run_it() below will be called
        # as part of the RunnablePassthrough infrastructure.
        _: Other = await super().ainvoke(input, config, **kwargs)

        # Collect intercepted outputs to report back to the tracing infrastructure.
        outputs: Dict[str, Any] = self.get_intercepted_outputs()
        return outputs

    async def run_it(self, inputs: Input) -> Output:
        """
        Transform a single input into an output.

        Args:
            inputs: The input to the `Runnable`.

        Returns:
            The output of the `Runnable`.
        """
        raise NotImplementedError

    def prepare_runnable_config(self, session_id: str = None,
                                callbacks: List[BaseCallbackHandler] = None,
                                recursion_limit: int = None,
                                use_run_name: bool = False) -> Dict[str, Any]:
        """
        Prepare a RunnableConfig for a Runnable invocation.  See:
        https://python.langchain.com/api_reference/core/runnables/langchain_core.runnables.config.RunnableConfig.html

        :param session_id: An id for the run
        :param callbacks: A list of BaseCallbackHandlers to use for the run
        :param recursion_limit: Maximum number of times a call can recurse.
        :return: A dictionary to be used for a Runnable's invoke config.
        """
        agent_name: str = self.invocation_context.get_agent_name()

        # Set up a run name for tracing purposes
        run_name: str = None
        if use_run_name:

            agent_prefix: str = ""
            if agent_name:
                agent_prefix = agent_name

            origin_name: str = ""
            if self.origin:
                full_name: str = Origination.get_full_name_from_origin(self.origin)
                origin_name = f"{full_name}"

            delimiter: str = ""
            if agent_prefix and origin_name:
                delimiter = ":"

            run_name: str = f"{agent_prefix}{delimiter}{origin_name}"

        runnable_config: Dict[str, Any] = {}

        # Add some optional stuff
        if session_id:
            runnable_config["configurable"] = {
                "session_id": session_id
            }

        if run_name:
            runnable_config["run_name"] = run_name

        if callbacks:
            runnable_config["callbacks"] = callbacks

        if recursion_limit:
            runnable_config["recursion_limit"] = recursion_limit

        # Only add metadata if we have something
        runnable_metadata: Dict[str, Any] = self.prepare_tracing_metadata()
        if runnable_metadata:
            runnable_config["metadata"] = runnable_metadata

        return runnable_config

    def prepare_tracing_metadata(self) -> Dict[str, Any]:
        """
        Prepare a dictionary of metadata for tracing purposes.

        :return: A dictionary of metadata for run tracing
        """
        runnable_metadata: Dict[str, Any] = {}

        # Add values for listed env vars if they have values.
        # Defaults are standard env vars for kubernetes deployments
        env_vars_str: str = os.getenv("AGENT_TRACING_METADATA_ENV_VARS",
                                      "POD_NAME POD_NAMESPACE POD_IP NODE_NAME")
        to_add: Dict[str, Any] = MetadataUtil.minimize_metadata(os.environ, env_vars_str)
        runnable_metadata.update(to_add)

        request_keys: str = os.getenv("AGENT_TRACING_METADATA_REQUEST_KEYS",
                                      os.getenv("AGENT_USAGE_LOGGER_METADATA",
                                                os.getenv("AGENT_FORWARDED_REQUEST_METADATA",
                                                          "request_id user_id")))
        request_metadata: Dict[str, Any] = self.invocation_context.get_metadata()
        to_add: Dict[str, Any] = MetadataUtil.minimize_metadata(request_metadata, request_keys)
        runnable_metadata.update(to_add)

        return runnable_metadata

    def get_intercepted_outputs(self) -> Dict[str, Any]:
        """
        :return: the intercepted outputs
        """
        intercepted_messages: List[BaseMessage] = self.interceptor.get_messages()

        messages: List[Dict[str, Any]] = messages_to_dict(intercepted_messages)
        outputs: Dict[str, Any] = {
            "messages": messages
        }
        return outputs




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict
from typing import Sequence

import copy
import logging
import pathlib

from leaf_server_common.logging.logging_setup import setup_logging

from neuro_san.service.http.logging.log_context_filter import LogContextFilter
from neuro_san.service.interfaces.event_loop_logger import EventLoopLogger


class HttpLogger(EventLoopLogger):
    """
    Custom logger class for use by Http server.
    """

    HTTP_LOGGER_NAME: str = "HttpServer"

    def __init__(self, forwarded_metadata: Sequence[str],
                 logging_config: Dict[str, Any]):
        """
        Constructor

        :param forwarded_metadata: list of metadata keys to be forwarded
        :param logging_config: logging configuration dictionary
        """
        # Initialize minimal metadata dictionary to contain some value
        # for each metadata key we expect to be used.
        self.base_metadata: Dict[str, Any] = {}
        for key in forwarded_metadata:
            self.base_metadata[key] = "None"
        self.base_metadata["source"] = HttpLogger.HTTP_LOGGER_NAME
        LogContextFilter.set_log_context()
        self.setup_logging(logging_config)
        # For our Http server, we have separate logging setup,
        # because things like "user_id" and "request_id"
        # can only be extracted and used on per-request basis.
        # Async Http server is single-threaded.
        self.logger = logging.getLogger(HttpLogger.HTTP_LOGGER_NAME)

    def info(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Info" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        self.prepare_filter(metadata)
        self.logger.info(msg, *args)

    def warning(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Warning" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        self.prepare_filter(metadata)
        self.logger.warning(msg, *args)

    def debug(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Debug" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        self.prepare_filter(metadata)
        self.logger.debug(msg, *args)

    def error(self, metadata: Dict[str, Any], msg: str, *args):
        """
        "Error" logging method.
        Prepare logger filter with request-specific metadata
        and delegate logging to underlying standard Logger.
        """
        self.prepare_filter(metadata)
        self.logger.error(msg, *args)

    def setup_logging(self, logging_config: Dict[str, Any]):
        """
        Setup logging from configuration file.

        :param logging_config: logging configuration dictionary
        """
        # Need to initialize the forwarded metadata default values before our first
        # call to a logger.
        current_dir: str = pathlib.Path(__file__).parent.parent.resolve()
        setup_logging(HttpLogger.HTTP_LOGGER_NAME,
                      default_log_dir=current_dir,
                      log_level_env="AGENT_SERVICE_LOG_LEVEL",
                      extra_logging_fields_defaults=self.base_metadata,
                      logging_config=logging_config)

        # This module within openai library can be quite chatty w/rt http requests
        logging.getLogger("httpx").setLevel(logging.WARNING)

    def prepare_filter(self, metadata: Dict[str, Any]):
        """
        Prepare logging filter using request metadata,
        """
        use_metadata: Dict[str, Any] = copy.copy(self.base_metadata)
        use_metadata.update(metadata)
        LogContextFilter.log_context.set(use_metadata)




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import List

from langchain_core.messages.base import BaseMessage

from neuro_san.internals.run_context.interfaces.run import Run
from neuro_san.internals.run_context.interfaces.tool_call import ToolCall
from neuro_san.internals.run_context.langchain.core.langchain_tool_call import LangChainToolCall


class LangChainRun(Run):
    """
    A LangChain implementation of a Run.
    """

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-positional-arguments
    def __init__(self, run_id_base: str, chat_history: List[BaseMessage], tool_name: str = None,
                 args: Any = None, tool_message: BaseMessage = None, invocation: str = None):
        """
        Constructor

        :param run_id_base: The basis string for run ids.
        :param chat_history: The chat history so far
        :param tool_name: The name of the tool to call
        :param args: The arguments used for the tool during the run
                    represented by this instance.
        :param tool_message: The most recent tool message produced during the run
                             represented by this instance.
        :param invocation: String describing how the tool wants to be invoked.
                            Can be: "chatbot" - implies waiting for an answer.
                                    "event" - implies no answer needed
                                    None - implies chatbot
        """
        self.id: str = self._create_run_id(run_id_base, chat_history)
        self.chat_history: List[BaseMessage] = chat_history
        self.tool_name: str = tool_name
        self.args: Any = args
        self.tool_message: BaseMessage = tool_message
        self.invocation: str = invocation

    @staticmethod
    def _create_run_id(run_id_base: str, chat_history: List[BaseMessage]) -> str:
        """
        :return: A run id string based on the uuid in run_id_base and
                 the length of the chat history so far.
        """
        chat_length = len(chat_history)

        # Zero pad the chat_length out to 4 digits.
        run_id: str = f"run_{run_id_base}_{chat_length:04}"
        return run_id

    def get_id(self) -> str:
        """
        :return: The string id of this run
        """
        return self.id

    def requires_action(self) -> bool:
        """
        :return: True if the status of the run requires external action.
                 False otherwise
        """
        # This allows the higher level check in branch tools
        # to know that we are doing something.
        return False

    def get_tool_calls(self) -> List[ToolCall]:
        """
        :return: A list of ToolCalls.
        """
        # At least return an empty list
        tool_calls: List[ToolCall] = []
        if self.args is not None:
            # We don't yet know how to get the list of tools an AgentExecutor actually
            # wants to call (as opposed to the list it *can( call, which is in the spec).
            # It probably has something to do with adding a CallbackManagerForChainRun
            # which somehow would get its on_agent_action() called when a tool is chosen in
            # AgentExecutor._perform_agent_action().  We'd probably need to set that up
            # and allow it to add tools/args to this run instance.
            #
            # For now, we just do a single tool given the run...
            # Create a ToolCall with the correct args and add that to the ToolCalls list
            tool_call: LangChainToolCall = LangChainToolCall(self.tool_name, self.args, self.id,
                                                             invocation=self.invocation)
            tool_calls.append(tool_call)

        return tool_calls

    def model_dump_json(self) -> str:
        """
        :return: This object as a JSON string
        """
        return "{ " + f"'id': '{self.id}'" + " }"

    def get_chat_history(self) -> List[BaseMessage]:
        """
        Get the most recent chat history.
        :return: The list of messages comprising the chat history
        """
        return self.chat_history

    def get_tool_message(self) -> BaseMessage:
        """
        This is used when the LangChainOpenAIFunctionTool wants to
        know what the most recent message was from the tool as
        its output.
        :return: The tool message comprising the tool output
        """
        return self.tool_message




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
from typing import Any
from typing import Dict


class AgentNetworkInspector:
    """
    Interface describing an entity that can examine the configuration of the
    agent network spec.
    Having this interface breaks some circular dependencies.
    """

    def get_config(self) -> Dict[str, Any]:
        """
        :return: The entire config dictionary given to the instance.
        """
        raise NotImplementedError

    def get_agent_tool_spec(self, name: str) -> Dict[str, Any]:
        """
        :param name: The name of the agent tool to get out of the registry
        :return: The dictionary representing the spec registered agent
        """
        raise NotImplementedError

    def get_name_from_spec(self, agent_spec: Dict[str, Any]) -> str:
        """
        :param agent_spec: A single agent to register
        :return: The agent name as per the spec
        """
        raise NotImplementedError

    def find_front_man(self) -> str:
        """
        :return: A single tool name to use as the root of the chat agent.
                 This guy will be user facing.  If there are none or > 1,
                 an exception will be raised.
        """
        raise NotImplementedError




# Copyright © 2023-2026 Cognizant Technology Solutions Corp, www.cognizant.com.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# END COPYRIGHT
"""
See class comment for details
"""
from typing import Any
from typing import Dict

from neuro_san.service.generic.async_agent_service import AsyncAgentService
from neuro_san.service.http.handlers.base_request_handler import BaseRequestHandler


class ConnectivityHandler(BaseRequestHandler):
    """
    Handler class for neuro-san "connectivity" API call.
    """

    async def get(self, agent_name: str):
        """
        Implementation of GET request handler for "connectivity" API call.
        """
        metadata: Dict[str, Any] = self.get_metadata()
        service: AsyncAgentService = await self.get_service(agent_name, metadata)
        if service is None:
            return

        self.application.start_client_request(metadata, f"{agent_name}/connectivity")
        try:
            data: Dict[str, Any] = {}
            result_dict: Dict[str, Any] = await service.connectivity(data, metadata)

            # Return response to the HTTP client
            self.set_header("Content-Type", "application/json")
            self.write(result_dict)

        except Exception as exc:  # pylint: disable=broad-exception-caught
            self.process_exception(exc)
        finally:
            self.do_finish()
            self.application.finish_client_request(metadata, f"{agent_name}/connectivity")
